/*
** Module   : MajicAdminClient.java
** Abstract : MAJIC specific admin client extension plugin 
**
** Copyright (c) 2009-2012, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description----------------------
** 001 NVS 20090522   @42486 Created initial empty file.
** 002 NVS 20090529   @42494 Initial version of the Majic client plugin.
** 003 CA  20090602   @42515 When editing an account with extension data, if 
**                           the 'Has Extensions' check-box is toggled, reload 
**                           the extension data. When Employee is chosen, the 
**                           Person field (from user def dialog) is left 
**                           untouched and editable.
** 004 CA  20090604   @42581 Handle the created-date field. Disable the majic
**                           related tabs when 'has extensions' check-box is
**                           not selected. Added resetDialog - it will reset
**                           all fields to some default values. 
** 005 CA  20090609   @42617 Changed or fixed the layout for some fields.
** 006 CA  20090612   @42663 The employee/customer name and the description for
**                           printer/entry-point will be truncated and set in 
**                           a label.
** 007 NVS 20090803   @43493 Refactored the code and relocated panel creation  
**                           methods using border and group layouts.
** 008 NVS 20090804   @43519 Implemented account extension data printing related
**                           methods: getPreviewHeaders(), getPreviewRow().
** 009 NVS 20090805   @43527 Implemented account extension data printing related
**                           methods: getExtendedHeaders(), getExtendedRows(),
**                           fillExtendedRow().
** 010 NVS 20090805   @43530 Added missing Date Created display and fixed NPEs
**                           in extensions printing caused by undefined items.
** 011 NVS 20090806   @43545 Added implementation of AdminCustomMenu interface
**                           and company name editor.
** 012 NVS 20090914   @43896 Fixed an implementation flaw in the account edit
**                           operation, which used to lose created-date, 
**                           user-num and key extensions and trashed user-level
**                           value. The extensions container for the account
**                           being edited is retained till the Save operation.
** 013 CA  20090915   @43904 Fixed NPE in validateExtension - when adding a  
**                           new user or editing a user without extensions, a  
**                           new extension object must be built in 
**                           loadExtensionData.
** 014 NVS 20090930   @44076 Validation of input no longer tries to steal focus
**                           and error messages finally stay visible.
**                           Fixed multiple issues with the table browser.
** 015 NVS 20091007   @44121 Changed to allow the extension fields in the add
**                           or edit user dialog to be enabled or disabled 
**                           according to granular admin permissions.
** 016 CA  20091026   @44198 Fixed regression introduced by H015 (@44121), 
**                           caused by parameter rename in layoutAllWidgets.
** 017 SIY 20091116   @44331 Added printer maintenance page. 
** 018 SIY 20091204   @44477 Added function maintenance page. 
** 019 SIY 20091211   @44497 Added panel menu builder. 
** 020 SIY 20091214   @44503 Added function lookup for panel menu builder. 
**                           Added print support for printer, function and 
**                           panel menu maintenance.
** 021 SIY 20091215   @44509 Fixed incorrect enabling of account extensions 
**                           when they are not actually present. As a 
**                           consequence this leads to inability to change 
**                           password if there are no account extensions.
** 022 SIY 20100419   @44819 Implemented automatic adding of function into
**                           directory if it does not exists.
** 023 SIY 20100422   @44831 Added call to refreshFunctions().
** 024 SVL 20100211   @44582 Added user search capabilities and menu of
**                           reports.
** 025 SVL 20110313          Added the ability to add/edit/delete permit-only
**                           functions.
** 026 CS  20121031          Exposed custom client standalone ACL via hook to admin ACL
*/

package aero.timco.majic.admin.client;

import java.awt.Component;
import java.awt.Container;
import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.event.*;
import java.awt.print.PrinterException;
import java.io.Serializable;
import java.text.DateFormat;
import java.text.MessageFormat;
import java.text.NumberFormat;
import java.util.*;

import javax.swing.*;
import javax.swing.GroupLayout.ParallelGroup;
import javax.swing.GroupLayout.SequentialGroup;
import javax.swing.JTable.PrintMode;
import javax.swing.border.EmptyBorder;
import javax.swing.border.EtchedBorder;
import javax.swing.border.TitledBorder;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.table.TableModel;
import javax.swing.text.*;

import aero.timco.majic.admin.server.MajicExtDef;
import aero.timco.majic.admin.server.MajicExtExports;
import aero.timco.majic.admin.server.MenuDef;

import com.goldencode.p2j.admin.AccountExtValidationException;
import com.goldencode.p2j.admin.AdminExports;
import com.goldencode.p2j.admin.TaggedName;
import com.goldencode.p2j.admin.client.*;
import com.goldencode.p2j.net.RemoteObject;
import com.goldencode.p2j.util.date;

/**
 * Extends the admin client to provide MAJIC specific operations through the
 * additional client menu action items and manages the MAJIC specific custom
 * account extension data for user or process account creation, editing and
 * searching.
 */
public class MajicAdminClient
implements ClientAccountExtension, AdminCustomMenu
{
   /**
    * Comment for <code>ABCDEFGHIJKLMNOPQRSTUVWXYZ_ABCDEFGHIJKLMNOPQRSTUVWXYZ</code>
    */
   private static final String VALID_CHARS = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ-_. $1234567890";

   /** set the amount of records retrieved per page in the browse dialog */
   private static final int PAGE_SIZE = 20;
   
   /** number of rows the account extension takes on extended printing */
   private static final int EXTENDED_ROWS = 10;
   
   /** number of columns the account extension takes on extended printing */
   private static final int EXTENDED_COLS = 4;
   
   /** current {@link IAccountScreen} instance. */
   private IAccountScreen ias = null;
   
   /** current {@link AdminClient} instance. */
   private AdminClient client = null;

   /** currently executed dialog */
   private JDialog dialog = null;
   
   /** instance of the {@link MajicExtExports} interface*/
   public MajicExtExports majic = null;

   /** main tab panel */
   private JTabbedPane jtabPane = null;

   // extension-field widgets

   /** has-extensions check-box */
   private JCheckBox t1chkHe = null;

   /** user-code field */
   private JTextField t1fUc = null;
   
   /** employee-id field */
   private JTextField t1fEId = null;
   
   /** employee-name field */
   private JLabel t1lEName = null;
   
   /** customer-id field */
   private JTextField t1fCId = null;
   
   /** customer-name field */
   private JLabel t1lCName = null;

   /** created-date field */
   private JTextField t1fCd = null;

   /** run-before field */
   private JCheckBox  t2chkRb = null;

   /** auto-quit field */
   private JTextField t2fAq = null;

   /** user-help field */
   private JTextField t2fUh = null;

   /** entry-point id field */
   private JTextField t2fEpId = null;
   
   /** entry-point name field */
   private JLabel t2lEpName = null;
   
   /** printer-id field */
   private JTextField t3fPrId = null;
   
   /** printer-name field */
   private JLabel t3lPrName = null;

   /** default-file field */
   private JTextField t3fDf = null;
   
   /** file-viewer field */
   private JTextField t3fFv = null;

   /** file-editor field */
   private JTextField t3fFe = null;

   /** top-dir field */
   private JTextField t3fTd = null;

   /** browse-mask field */
   private JCheckBox  t4chkBm = null;

   /** user-level field */
   private JTextField t4fUl = null;

   /** valid-apps field */
   private JTextField t4fVa = null;

   /** */
   private JTextArea t5fDesc = null;

   // browse-dialog widgets
   
   /** main browse panel */
   private JPanel bwPanel = null;

   /** scroller's titled border */
   private TitledBorder bwScrollerTitle = null; 
 
   /** table to display browse results */
   private JTable bwTable = null;

   /** related table model */
   private TaggedDataTableModel bwBtm = null;

   /** related table selection model */
   private ListSelectionModel bwBsm = null;
   
   /** browse ID label */
   private JLabel bwlId = null;
   
   /** browse ID field */
   private JTextField bwfId = null;
   
   /** brows DESC label */
   private JLabel bwlD = null;
   
   /** browse DESC field */
   private JTextField bwfD = null;
      
   /** browse label with page #*/
   private JLabel bwlPg = null;

   // browse handlers for various fields 
   
   /** cust-num browse handler */
   private CustomerBrowser custBrowser = null;   
   
   /** emp-num browse handler */
   private EmployeeBrowser empBrowser = null;   
   
   /** printer browse handler */
   private PrinterBrowser printBrowser = null;
   
   /** entry-point handler */
   private EntryPointBrowser entryPointBrowser = null;
   
   /** all the extended components */
   private Component[] extComps = null;
   
   /** all the extended fields which are represented as integer fields */
   private Set<Component> intComps = null;

   /** currently browser handler */
   private IBrowser currentBrowser = null;
   
   /** current browser page */
   private int browsePage = 0;
   
   /** current browser ID filter */
   private String bwIdFilter = null;

   /** current browser NAME filter */
   private String bwNameFilter = null;

   /** field with current account name. when null, it means is in add mode. */
   private String accountName = null;
   
   /** flag indicating if data loading is executed */
   private boolean loadingData = false;
   
   /** custom menu position */
   private static final int MAJIC_MENU_POS = 3;

   private static final String SCREEN_PRINTER_MAINTENANCE = "printer_maint";

   private static final String TITLE_PRINTER_MAINTENANCE = "Printer Maintenance";

   private static final String SCREEN_FUNCTION_MAINTENANCE = "func_maint";

   private static final String TITLE_FUNCTION_MAINTENANCE = "System Function Maintenance";

   private static final String SCREEN_MENU_BUILDER = "menu_builder";

   private static final String TITLE_MENU_BUILDER = "Panel Menu Builder";
   
   /** custom MAJIC menu */
   private JMenu majicMenu = null;
   
   /** custom Functions menu item */
   private JMenuItem fun = null;

   /** custom Printer Maintenance menu item */
   private JMenuItem xprint = null;

   /** custom Panel Menu Builder menu item */
   private JMenuItem xmbuild = null;
   
   /** custom MAJIC Reports menu */
   private JMenu majicReportsMenu = null;
   
   /** custom User Authorization Form menu item*/
   private JMenuItem repUAFMenuItem = null;

   /** menu to add the minor menu item */
   private JMenu psr = null;
   
   /** position of the first added item */
   private int psrCount = 0;
   
   /** custom Company Name menu item */
   private JMenuItem comp = null;

   /** company name editor dialog */
   
   private JDialog dlg = null;
   
   /** account extensions container for the account being edited */
   private MajicExtDef extDef = null;

   /** Printers table model. */
   private TaggedDataTableModel printerTm = null;

   /** Functions table model. */
   private TaggedDataTableModel functionTm = null;

   /** Menus table model, left pane. */
   private TaggedDataTableModel menuListTm = null;

   /** Menus table model, right pane. */
   private TaggedDataTableModel menuItemsTm = null;

   /** Printer table. */
   private JTable printers = null;
   
   /** Function table. */
   private JTable functions = null;
   
   /** Menu table, left pane. */
   private JTable menus = null;

   /** Menu table, right pane. */
   private JTable menuItems = null;

   /** Local menu DB. */
   private Map<String, MenuDef> menuMap = new LinkedHashMap<String, MenuDef>();

   /** Function cache for menu editor. */
   private ArrayList<String> funcCache = null;
   
   /** The "User Authorization Form" report */
   private UserAuthFormReport repUserAuthForm = null;

   /** The "User List" report */
   private UserListReport repUserList = null;

   /** The "User Auth By Function" report */
   private UserAuthByFuncReport repUserAuthByFunc = null;

   /** Input field for user search requests. */
   private JTextField findUserInput = null;
  
   /**
    * Get the layout mode of the plugin:
    * <p>
    * {@link #LAYOUT_ALL_WIDGETS}: plugin is responsible to layout all widgets
    * <p>
    * {@link #LAYOUT_EXT_WIDGETS}: plugin is responsible to layout only ext
    * widgets
    *  
    * @return   {@link #LAYOUT_ALL_WIDGETS}
    */
   public int getLayoutMode()
   {
      return LAYOUT_ALL_WIDGETS;
   }
   
   /**
    * Reset all extension widgets to a predefined state.
    */
   public void resetDialog()
   {
      this.loadingData = true;
      
      // reset the account name
      this.accountName = null;

      resetAllExtFields();
      
      // enable the hasExtension check-box
      t1chkHe.setSelected(true);

      // init the user-code and created-date fields
      handleHasExtCheckbox();
      
      this.loadingData = false;
   }

   /**
    * Layout the extension fields; this method can re-layout the default 
    * widgets too, if needed. If <code>accountName</code> is provided, it
    * will populate the extension fields with its extension data, if any. 
    * <p>
    * For Majic, the layout is a tabbed panel with following components:
    * 
    * <ol>
    *    <li>Account Properties
    *       <ul>
    *          <li>account defaults (specific to user or process)
    *          </li>
    *          <li>check-box with 'Has extension data ?'</li>
    *          <li>account data (employee, customer, user code)</li>
    *       </ul>
    *    </li>
    *    <li>Application Defaults
    *       <ul>
    *          <li>menu defaults (entry point, auto-quit)</li>
    *          <li>application help defaults (user help, run before)<li>
    *       </ul>
    *    </li>
    *    <li>System Defaults
    *       <ul>
    *          <li>printer default (printer)</li>
    *          <li>system defaults (default file, file viewer, file editor, 
    *              top directory)
    *          </li>
    *       </ul>
    *    </li>
    *    <li>Security
    *       <ul>
    *          <li>security (browse mask, user-level, valid apps)</li>
    *       </ul>
    *    </li>
    *    <li>Description</li>
    * </ol>
    * 
    * The following fields will have a button which in turn opens a search
    * dialog: employee, customer, printer.
    *  
    * @param    userName
    *           The account for which extension field layout is needed. If 
    *           <code>null</code>, it means a new account is added.
    * 
    * @param    writable
    *           <code>true</code> if this is for enabled inputs.
    *           <code>false</code> for read-only panel
    *
    * @return   the root panel which will be shown as a modal dialog.
    */
   public JPanel layoutAllWidgets(String userName, boolean writable)
   {
      final String[] tabs = 
      { 
            "Account", 
            "Application Defaults",
            "System Defaults",
            "Security",
            "Description"
      };
      
      // tab 1 - account
      JPanel t1pU  = buildAccountDefaultsPanel();
      // remove buttons and build the cancel/save panel
      
      final JButton cb = ias.getCloseButton();
      JButton cs = ias.getSaveButton();

      JPanel pnlb = AdminUtils.groupComponents(
            new Component[][] { { cb, new JLabel(""), cs } });

      JPanel t1pUD = buildAccountDataPanel();
      
      // tab 2 - application
      JPanel t2pM = buildMenuDefaults();
      JPanel t2pH = buildAppHelpDefaults();
      
      // tab 3 - system
      JPanel t3pP = buildPrinterDefaults();
      JPanel t3pS = buildSystemDefaults();
      
      // tab 4 - security
      JPanel t4pP = buildSecurity();
      
      // tab 5 - description
      JPanel t5pD = buildDescription();

      // build layout for each tab
      final JPanel[][] tabsDef = 
      {
            { t1pU, t1pUD },
            { t2pM, t2pH },
            { t3pP, t3pS },
            { t4pP },
            { t5pD }
      };
      
      jtabPane = new JTabbedPane();
      
      // build the actual tabs
      for (int i = 0; i < tabsDef.length; i++)
      {
         // build a border'd panel with vertical components
         JPanel pnl = AdminUtils.borderComponents(tabsDef[i], true);
         
         jtabPane.addTab(tabs[i], pnl);
      }
      
      // add the tabbed pane to a panel
      JPanel pnl = new JPanel();
      pnl.setLayout(new BorderLayout());
      
      pnl.add(jtabPane, BorderLayout.NORTH);
      pnl.add(pnlb, BorderLayout.CENTER);
      
      // set any needed handlers
      // employee text field
      t1fEId.addActionListener(new ActionListener() 
      {
         /**
          * Event handling method.
          * 
          * @param    evt
          *           Action event.
          */
         public void actionPerformed(ActionEvent evt) 
         {
            handleEmployeeBrowse();
         }
      });

      // customer text field
      t1fCId.addActionListener(new ActionListener() 
      {
         /**
          * Event handling method.
          * 
          * @param    evt
          *           Action event.
          */
         public void actionPerformed(ActionEvent evt) 
         {
            handleCustomerBrowse();
         }
      });
      
      // entry-point text field
      t2fEpId.addActionListener(new ActionListener() 
      {
         /**
          * Event handling method.
          * 
          * @param    evt
          *           Action event.
          */
         public void actionPerformed(ActionEvent evt) 
         {
            handleEntryPointBrowse();
         }
      });
      
      // printer text field
      t3fPrId.addActionListener(new ActionListener() 
      {
         /**
          * Event handling method.
          * 
          * @param    evt
          *           Action event.
          */
         public void actionPerformed(ActionEvent evt) 
         {
            handlePrinterBrowse();
         }
      });
      
      // has ext check-box    
      t1chkHe.addActionListener(new ActionListener() 
      {
         /**
          * Event handling method.
          * 
          * @param    evt
          *           Action event.
          */
         public void actionPerformed(ActionEvent evt) 
         {
            handleHasExtCheckbox();
         }
      });
      
      // focus-lost for emp-num field       
      t1fEId.addFocusListener(new FocusListener() 
      {
         public void focusLost(FocusEvent evt) 
         {
            // ignore temporary events
            if (evt.isTemporary())
            {
               return;
            }
            
            validateEmployee();
         }

         public void focusGained(FocusEvent evt) 
         {
            // no-op
         }
      });
      
      // focus-lost for cust-num field       
      t1fCId.addFocusListener(new FocusListener() 
      {
         public void focusLost(FocusEvent evt) 
         {
            // ignore temporary events
            if (evt.isTemporary())
            {
               return;
            }
            
            validateCustomer();
         }
         
         public void focusGained(FocusEvent evt) 
         {
            // no-op
         }
      });
      
      // focus-lost for entry-point field       
      t2fEpId.addFocusListener(new FocusListener() 
      {
         public void focusLost(FocusEvent evt) 
         {
            validateEntryPoint();
         }
         
         public void focusGained(FocusEvent evt) 
         {
            // no-op
         }
      });
      
      // focus-lost for printer name field       
      t3fPrId.addFocusListener(new FocusListener() 
      {
         public void focusLost(FocusEvent evt) 
         {
            validatePrinter();
         }
         
         public void focusGained(FocusEvent evt) 
         {
            // no-op
         }
      });
      
      // build a list of extension fields, which can be disabled
      extComps = new Component[] 
      { 
         t1fUc, t1fEId, t1fCId, t1fCd, 
         t2chkRb, t2fAq, t2fUh, t2fEpId, 
         t3fPrId, t3fDf, t3fFv, t3fFe, t3fTd, 
         t4chkBm, t4fUl, t4fVa, 
         t5fDesc 
      };

      // build a list of components which require integer-type validation
      intComps = new HashSet<Component>();
      intComps.add(t2fAq); // auto-quit
      intComps.add(t4fUl); // user-level
      intComps.add(t1fCId); // cust-num
      intComps.add(t1fEId); // emp-num
      
      loadExtensionData(userName, false);
      
      // conditionally enable/disable inputs
      t1chkHe.setEnabled(writable);
      
      for (int i = 0; i < extComps.length; i++)
      {
         extComps[i].setEnabled(writable);
      }
      
      return pnl;
   }
   
   /**
    * Validate the employee id, on focus lost.
    * 
    * @return <code>true</code> if validation succeeded
    */
   private boolean validateEmployee()
   {
      if ("?".equals(t1fEId.getText()))
      {
         return true;
      }
      
      boolean valid = true;
      String empName = "";
      
      try
      {
         int empno = validateIntegerField(t1fEId, "invalid employee number");
         
         if (empno != 0)
         {
            AdminUtils.busy(client);
            empName = majic.getEmployeeName(empno);
            AdminUtils.idle(client);
            
            if (empName == null)
            {
               client.beep("employee doesn't exist");
               valid = false;
               empName = "";
            }
         }
         
         t1lEName.setText(truncateText(empName, true));

         // set the person name
         if (empBrowser == null)
         {
            empBrowser = new EmployeeBrowser();
         }
         empBrowser.handleChoose(t1fEId.getText(), empName);
      }
      catch (AccountExtValidationException e) 
      {
         empName = "";
         t1lEName.setText(truncateText(empName, true));
         client.beep(e.getMessage());
         valid = false;
      }
      
      // clear error message on success
      if (valid)
      {
         client.ok();
      }

      return valid;
   }
   
   /**
    * Validate the customer id, on focus lost.
    * 
    * @return   <code>true</code> if validation succeeded
    */
   private boolean validateCustomer()
   {
      if ("?".equals(t1fCId.getText()))
      {
         return true;
      }

      boolean valid = true;
      String custName = "";
      
      try
      {
         int custno = validateIntegerField(t1fCId, "invalid customer number");
         
         if (custno != 0)
         {
            AdminUtils.busy(client);
            custName = majic.getCustomerName(custno);
            AdminUtils.idle(client);
            
            if (custName == null)
            {
               client.beep("customer doesn't exist");
               valid = false;
               custName = "";
            }
         }
         
         t1lCName.setText(truncateText(custName, true));

         // set the customer name
         if (custBrowser == null)
         {
            custBrowser = new CustomerBrowser();
         }
         custBrowser.handleChoose(t1fCId.getText(), custName);
      }
      catch (AccountExtValidationException e) 
      {
         custName = "";
         t1lCName.setText(truncateText(custName, true));
         client.beep(e.getMessage());
         valid = false;
      }
      
      // clear error message on success
      if (valid)
      {
         client.ok();
      }

      return valid;
   }
   
   /**
    * Validate the entry-point, on focus lost.
    * 
    * @return   <code>true</code> if validation succeeded
    */
   private boolean validateEntryPoint()
   {
      if ("?".equals(t2fEpId.getText()))
      {
         return true;
      }

      boolean valid = true;
      
      String entryPoint = t2fEpId.getText().trim();
      String desc = "";
      
      if (entryPoint.length() > 0)
      {
         AdminUtils.busy(client);
         desc = majic.getEntryPointDescription(entryPoint);
         AdminUtils.idle(client);
         
         if (desc == null)
         {
            client.beep("entry-point doesn't exist");
            valid = false;
            
            desc = "";
         }
      }
      
      t2lEpName.setText(truncateText(desc, false));

      // clear error message on success
      if (valid)
      {
         client.ok();         
      }

      return valid;
   }
   
   /**
    * Validate the printer, on focus lost.
    * 
    * @return   <code>true</code> if validation succeeded
    */
   private boolean validatePrinter()
   {
      if ("?".equals(t3fPrId.getText()))
      {
         return true;
      }

      boolean valid = true;
      
      String printer = t3fPrId.getText().trim();
      String type = "";
      
      if (printer.length() > 0)
      {
         AdminUtils.busy(client);
         type = majic.getPrinterType(printer);
         AdminUtils.idle(client);
         
         if (type == null)
         {
            client.beep("printer doesn't exist");
            valid = false;
            
            type = "";
         }
      }
      
      t3lPrName.setText(truncateText(type, false));

      // clear error message on success
      if (valid)
      {
         client.ok();         
      }

      return valid;
   }
   
   /**
    * If the specified account is non-null and has extensions, load the
    * extension data for this account. If it has no extensions, disable and
    * reset all fields.
    * 
    * @param    accountName
    *           the account for which data should be loaded
    * @param    fromCheckbox
    *           <code>true</code> if this call is generated by the check-box
    *           'has extensions'
    */
   private void loadExtensionData(String accountName, boolean fromCheckbox)
   {
      loadingData = true;
      this.accountName = accountName;

      if (accountName == null)
      {
         // in add mode, reset all components
         resetAllExtFields();

         // select has-extensions check-box
         t1chkHe.setSelected(true);
         
         handleHasExtCheckbox();

         loadingData = false;

         extDef = new MajicExtDef();

         return;
      }
      
      // an account name is provided, get extension data
      extDef = majic.getAccountExtensions(accountName);
      
      boolean enableCheckBox = (extDef != null);
      
      if (extDef != null)
      {
         // has extension data, load it

         // employee
         t1fEId.setText(extDef.empNum + "");
         t1lEName.setText(
            truncateText(majic.getEmployeeName(extDef.empNum), true));

         // customer
         t1fCId.setText(extDef.custNum + "");
         t1lCName.setText(
            truncateText(majic.getCustomerName(extDef.custNum), true));
         
         // user-code
         t1fUc.setText(extDef.userCode);
         
         // date created
         t1fCd.setText(extDef.createdDate != null ? 
                       extDef.createdDate.toString() : "");
         
         // entry point
         t2fEpId.setText(extDef.entryPoint);
         t2lEpName.setText(
             truncateText(majic.getEntryPointDescription(extDef.entryPoint), 
                          false));
         
         // auto-quit
         t2fAq.setText("" + extDef.autoQuit);
         
         // user-help
         t2fUh.setText(extDef.userHelp);
         
         // run-before
         t2chkRb.setSelected(extDef.runBefore);
         
         // printer
         t3fPrId.setText(extDef.printer);
         t3lPrName.setText(
            truncateText(majic.getPrinterType(extDef.printer), true));
         
         // default file
         t3fDf.setText(extDef.defaultFile);
         
         // file viewer
         t3fFv.setText(extDef.fileViewer);
         
         // file editor
         t3fFe.setText(extDef.fileEditor);
         
         // top dir
         t3fTd.setText(extDef.topDir);
         
         // browse mask
         t4chkBm.setSelected(extDef.browseMask);
         
         // user-level
         t4fUl.setText(extDef.userLevel + "");
         
         // valid apps
         t4fVa.setText(extDef.validApps);
         
         // description
         t5fDesc.setText(extDef.description);
      }
      else
      {
         extDef = new MajicExtDef();
      }

      if (!fromCheckbox)
      {
         t1chkHe.setSelected(enableCheckBox);

         // enable/disable extension fields, depending if there are extensions
         handleHasExtCheckbox();
      }

      loadingData = false;
   }

   /**
    * Truncate the text to a predefined length.
    * 
    * @param    text
    *           The text which needs to be truncated.
    * @param    firstTab
    *           <code>true</code> if the text is to be displayed in the first
    *           tab.
    * 
    * @return   the truncated text.
    */
   private String truncateText(String text, boolean firstTab)
   {
      if (text == null || text.trim().length() == 0)
      {
         return "";
      }
      
      text = text.trim();
      int length = firstTab ? 40 : 45;

      if (text.length() > length)
      {
         text = text.substring(0, length);
      }
      
      return " " + text;
   }
   
   /**
    * Show the search dialog backed by the specified browser
    * 
    * @param    browser
    *           the browser instance used to retrieve data
    */
   private void showSearchDialog(IBrowser browser)
   {
      // check if browser was requested
      if (!"?".equals(browser.getIdField().getText()))
      {
         return;
      }
      
      currentBrowser = browser;
      
      // build dialog, if not already built
      if (bwPanel == null)
      {
         buildBrowseDialog();
      }

      // reset filters
      bwfId.setText("");
      bwfD.setText("");
      
      // set the model for the table
      bwBtm = new TaggedDataTableModel(currentBrowser.getFirstColLabel(), 
                                       currentBrowser.getSecondColLabel());
      bwBtm.setSort(false);
      bwTable.setModel(bwBtm);
      bwBsm = bwTable.getSelectionModel();

      // reset page index
      browsePage = 0;
      
      // reset browse filters
      bwIdFilter = bwNameFilter = null;
      
      // display modal dialog
      dialog = AdminUtils.modalDialog(client, 
                                      currentBrowser.getTitle(), 
                                      bwPanel);
      
      bwScrollerTitle.setTitle(currentBrowser.getTitle());
      // change table columns
      bwBtm.setColumnName(0, currentBrowser.getFirstColLabel());
      bwBtm.setColumnName(1, currentBrowser.getSecondColLabel());
      bwTable.repaint();
      // change filter label texts
      bwlId.setText(currentBrowser.getFirstColLabel());
      bwlD.setText(currentBrowser.getSecondColLabel());
      
      refreshBrowser();
      
      dialog.setVisible(true);

      if (dialog != null)
      {
         // dialog dismissed by closing window
         handleCancel();
      }
   }

   /**
    * Build the browser dialog. This is a TaggedName backed table, with
    * pagination and filters for name and tag.
    */
   private void buildBrowseDialog()
   {      
      bwTable = new JTable();
      
      JScrollPane bwScroller = new JScrollPane();
      bwScroller.getViewport().add(bwTable);
      bwScrollerTitle = new TitledBorder
         (new EtchedBorder(EtchedBorder.RAISED), "");
      bwScroller.setBorder(bwScrollerTitle);
      
      bwScroller.setViewportBorder(new EmptyBorder(5, 5, 5, 5));

      // create the pagination widgets
      JButton bwbP = new JButton("Prev page");
      bwlPg = new JLabel("Page #");
      bwlPg.setHorizontalAlignment(JLabel.CENTER);
      JButton bwbN = new JButton("Next page");
      
      // add the components with border layout, horizontally
      JPanel bwPagPanel = AdminUtils.borderComponents(
            new Component[] 
               { 
                  bwbP, bwlPg, bwbN
               },
            false
      );

      // build the north panel with the pagination and table
      JPanel bwNorthPanel = AdminUtils.groupComponents(
            new Component[][] 
               { 
                  { bwPagPanel },
                  { bwScroller }
               }
      );
      
      // creating control widgets
      bwlId = new JLabel(" ID ");      
      bwfId = new JTextField(13);      
      bwlD = new JLabel(" Description ");
      bwfD = new JTextField(13);      
      JButton bwbF = new JButton("Filter");
      
      JPanel bwCtlPanel = AdminUtils.groupComponents(
            new Component[][] 
               { 
                  { bwlId, bwfId, bwlD, bwfD, bwbF }
               }, 
            " Filter "
      );
      
      // add cancel/save buttons
      JButton bwbC = new JButton("Cancel");
      JButton bwbS = new JButton("Choose");

      JPanel pnl = AdminUtils.borderComponents(
            new Component[] { bwbC, new JLabel(""), bwbS }, false);
      
      bwPanel = AdminUtils.groupComponents(
            new Component[][] 
                { 
                  { bwNorthPanel }, 
                  { bwCtlPanel }, 
                  { pnl }
                }
      );

      // set handlers
      // filter button
      bwbF.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused")
         public void actionPerformed(ActionEvent evt) 
         {
            handleBrowseFilter();
         }
      });

      // cancel button
      bwbC.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused")
         public void actionPerformed(ActionEvent evt) 
         {
            handleCancel();
         }
      });
      
      // choose button
      bwbS.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused")
         public void actionPerformed(ActionEvent evt) 
         {
            handleBrowseChoose();
         }
      });
      
      // previous page button
      bwbP.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused")
         public void actionPerformed(ActionEvent evt) 
         {
            handleBrowsePage(browsePage - 1);
         }
      });
      
      // next page button
      bwbN.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused")
         public void actionPerformed(ActionEvent evt) 
         {
            handleBrowsePage(browsePage + 1);
         }
      });
   }

   /**
    * Show the records in the specified page.
    * 
    * @param    page
    *           page where navigation should be tried
    */
   private void handleBrowsePage(int page)
   {
      if (page < 0)
      {
         client.beep("you are at the first page");
         return;
      }
      
      // attempt to retrieve records in the specified page
      int oldPage = browsePage;
      browsePage = page;
      
      boolean rc = refreshBrowser();
      
      // attempt failed, restore page number
      if (!rc)
      {
         browsePage = oldPage;
      }
   }
   
   /**
    * Handler for Choose button press in the browse dialog.
    */
   private void handleBrowseChoose()
   {
      if (bwTable.getSelectedRows().length != 1)
      {
         client.beep("select only one row");
         return;
      }

      int row = bwTable.getSelectedRow();
      if (row < 0)
      {
         client.beep("no row selected");
         return;
      }

      // populate center panel's groups list
      String id = (String) bwBtm.getValueAt(row, 0);
      String desc = (String) bwBtm.getValueAt(row, 1);

      currentBrowser.getIdField().setText(id);
      currentBrowser.getDescField().setText(desc);

      // handle any custom actions after a row was selected in the browse
      currentBrowser.handleChoose(id, desc);
      
      // hide the dialog
      dialog.setVisible(false);
      dialog = null;
   }
   
   /**
    * Handle Cancel button press in the browse dialog.
    */
   private void handleCancel()
   {
      // clear error messages
      client.ok();
      
      // hide the dialog
      dialog.setVisible(false);
      dialog = null;
   }
   
   /**
    * Handler filter button press in the browse dialog.
    */
   private void handleBrowseFilter()
   {
      // get the filters
      bwIdFilter = bwfId.getText().trim();
      bwNameFilter = bwfD.getText().trim();
      
      // reset page number
      browsePage = 0;
      
      refreshBrowser();
   }

   /**
    * Refresh browser data.
    * 
    * @return   <code>true</code> if data was received from the server.
    */
   private boolean refreshBrowser()
   {
      if (browsePage == 0)
      {
         bwBsm.clearSelection();
         bwBtm.clear();
      }

      // get the list of groups from the server
      AdminUtils.busy(client);
      TaggedName[] list = currentBrowser.findRecords(bwIdFilter, 
                                                     bwNameFilter, 
                                                     browsePage, 
                                                     PAGE_SIZE);
      AdminUtils.idle(client);
      
      if (list != null)
      {
         if (list.length > 0)
         {
            // the list may be only partially filled with the null fillers
            List<TaggedName> ltn = new LinkedList<TaggedName>();
            
            for (int i = 0; i < list.length; i ++)
            {
               if (list[i] == null)
               {
                  break;
               }
               ltn.add(list[i]);
            }
            
            if (ltn.size() < PAGE_SIZE)
            {
               list = ltn.toArray(new TaggedName[]{});
            }
            
            if (list.length > 0)
            {
               bwBsm.clearSelection();
               bwBtm.loadAll(list);
               
               // attempt succeeded, set the page label
               bwlPg.setText("Page #" + (browsePage + 1));
            }
            else
            {
               if (browsePage > 0)
               {
                 client.beep("you are at the last page");
               }
               else
               {
                 client.beep("no data");
               }
            }  
         }
      }
      else
      {
         if (browsePage > 0)
         {
           client.beep("you are at the last page");
         }
         else
         {
           client.beep("no data");
         }
      }
      
      return list != null && list.length > 0;
   }
   
   /**
    * Handle employee browse. When pressing '?' in the employee-id field, it
    * will show the search dialog.
    */
   private void handleEmployeeBrowse()
   {
      if (empBrowser == null)
      {
         empBrowser = new EmployeeBrowser();
      }

      showSearchDialog(empBrowser);
   }

   /**
    * Handle customer browse. When pressing '?' in the customer-id field, it
    * will show the search dialog.
    */
   private void handleCustomerBrowse()
   {
      if (custBrowser == null)
      {
         custBrowser = new CustomerBrowser();
      }

      showSearchDialog(custBrowser);
   }
   
   /**
    * Handle entry-point browse. When pressing '?' in the entry-point name 
    * field, it will show the search dialog.
    */
   private void handleEntryPointBrowse()
   {
      if (entryPointBrowser == null)
      {
         entryPointBrowser = new EntryPointBrowser();
      }

      showSearchDialog(entryPointBrowser);
   }
   
   /**
    * Handle printer browse. When pressing '?' in the printer-id field, it
    * will show the search dialog.
    */
   private void handlePrinterBrowse()
   {
      if (printBrowser == null)
      {
         printBrowser = new PrinterBrowser();
      }

      showSearchDialog(printBrowser);
   }

   /**
    * Handle the has-extension check-box. When ON, all extension-related
    * widgets will be enabled; also, the Person field will be disabled. When
    * OFF, all extension widgets are disabled and reset; the Person field will
    * be enabled.
    */
   private void handleHasExtCheckbox()
   {
      boolean hasExt = t1chkHe.isSelected();

      // enable/disable all tabs except first one
      for (int i = 1; i < jtabPane.getComponentCount(); i++)
      {
         jtabPane.setEnabledAt(i, hasExt);
      }
      
      for (int i = 0; i < extComps.length; i++)
      {
         extComps[i].setEnabled(hasExt);
      }
      
      if (!hasExt)
      {
         resetAllExtFields();         
      }

      // enable PERSON field, if this is the user-account dialog
      /*
      Component c = ias.getWidgetByName(AdminUserAccounts.WIDGET_PERSON);
      if (c != null)
      {
         // found the person widget, enable it only if no extensions are 
         // needed
         c.setEnabled(!hasExt || "0".equals(t1fEId.getText()));
      }
      */
      
      // disable the user-code and create-date fields
      t1fUc.setEnabled(false);
      t1fCd.setEnabled(false);
      
      if (hasExt && majic.getAccountExtensions(accountName) == null)
      {
         // only if it has extensions and there are no existing ones
         // get the user-code value
         t1fUc.setText(majic.getNextUserCode());
         
         // get the created date too
         date d = new date(majic.getCreatedDate().getDate());
         t1fCd.setText(d.toString());
      }

      // reload its data only if this is not a new account 
      // (to avoid recursion)
      if (!loadingData && hasExt)
      {
         loadExtensionData(accountName, true);
      }
   }

   /**
    * Reset all the extension-related fields.
    */
   private void resetAllExtFields()
   {
      for (int i = 0; i < extComps.length; i++)
      {
         // reset the field
         if (extComps[i] instanceof JTextComponent)
         {
            String text = intComps.contains(extComps[i]) ? "0" : "";
            ((JTextComponent) extComps[i]).setText(text);
         }
      
         if (extComps[i] instanceof AbstractButton)
         {
            ((AbstractButton) extComps[i]).setSelected(false);
         }
      }
      
      // reset text of other additional widgets
      t1lCName.setText("");
      t1lEName.setText("");
      t2lEpName.setText("");
      t3lPrName.setText("");
   }

   /**
    * Build the description panel.
    * 
    * @return   a panel with description widgets.
    */
   private JPanel buildDescription()
   {
      t5fDesc = new JTextArea();
      
      t5fDesc.setWrapStyleWord(true);
      
      JScrollPane pnl = new JScrollPane();
      pnl.setBorder(new EtchedBorder(EtchedBorder.RAISED));
      pnl.getViewport().add(t5fDesc);

      // build a matrix with the layout
      final Component[][] comps = 
      {
         { new JLabel(" Description ") }, 
         { pnl },
      };

      return AdminUtils.groupComponents(comps, null);
   }

   /**
    * Build the security panel. The following widgets are here: browse-mask,
    * user-level, valid-apps.
    * 
    * @return   a panel with security widgets
    */
   private JPanel buildSecurity()
   {
      t4chkBm = new JCheckBox();
      t4fUl = new JTextField(3);
      t4fVa = new JTextField(20);

      t4fUl.setMaximumSize(t4fUl.getPreferredSize());

      JPanel pnl = AdminUtils.groupComponents(
        new Component[][] { { t4chkBm, new JLabel(" User Level "), t4fUl } });
      // remove auto-gapings from the panel
      GroupLayout layout = (GroupLayout) pnl.getLayout();
      layout.setAutoCreateContainerGaps(false);
      layout.setAutoCreateGaps(false);
      
      // build a matrix with the layout
      final Component[][] comps = 
      {
         { new JLabel(" Browse Mask "), pnl },
         { new JLabel(" Valid Apps "), t4fVa },
      };

      return AdminUtils.groupComponents(comps, "Security");
   }

   /**
    * Build the system-defaults panel. The following widgets are here:
    * default- file, file-viewer, file-editor, top-directory.
    * 
    * @return   a panel with system-defaults widgets
    */
   private JPanel buildSystemDefaults()
   {
      t3fDf = new JTextField(20);
      t3fFv = new JTextField(20);
      t3fFe = new JTextField(20);
      t3fTd = new JTextField(20);

      // build a matrix with the layout
      final Component[][] comps = 
      {
         { new JLabel(" Default File "), t3fDf },
         { new JLabel(" File Viewer "), t3fFv },
         { new JLabel(" File Editor "), t3fFe },
         { new JLabel(" Top Directory "), t3fTd },
      };

      return AdminUtils.groupComponents(comps, "System Defaults");
   }

   /**
    * Build the printer panel. The following widgets are here: printer id and
    * printer name.
    * 
    * @return   a panel with printer widgets
    */
   private JPanel buildPrinterDefaults()
   {
      t3fPrId = new JTextField(8);
      t3fPrId.setToolTipText(
         "Enter ? and press RETURN to search for printers");
      t3fPrId.setMaximumSize(t3fPrId.getPreferredSize());

      t3lPrName = new JLabel();
      JPanel pnl = AdminUtils.groupComponents(
            new Component[][] { { t3fPrId, t3lPrName } });

      // remove auto-gapings from the panel
      GroupLayout layout = (GroupLayout) pnl.getLayout();
      layout.setAutoCreateContainerGaps(false);
      layout.setAutoCreateGaps(false);
      
      // build a matrix with the layout
      final Component[][] comps = 
      {
         { new JLabel(" Printer "), pnl },
      };

      return AdminUtils.groupComponents(comps, "Printer Defaults");
   }

   /**
    * Build the app-help panel. The following widgets are here: user-help and
    * run-before.
    * 
    * @return   a panel with app-help widgets
    */
   private JPanel buildAppHelpDefaults()
   {
      t2fUh = new JTextField(20);
      t2chkRb = new JCheckBox();

      // t2fUh.setMaximumSize(t2fUh.getPreferredSize());

      // build a matrix with the layout
      final Component[][] comps = 
      {
         { new JLabel(" User Help "), t2fUh },
         { new JLabel(" Run Before "), t2chkRb },
      };

      return AdminUtils.groupComponents(comps, "Help Defaults");
   }

   /**
    * Build the menu-defaults panel. The following widgets are here: entry-
    * point and auto-quit.
    * 
    * @return   a panel with menu-defaults widgets
    */
   private JPanel buildMenuDefaults()
   {
      t2fEpId = new JTextField(8);
      t2fEpId.setToolTipText(
         "Enter ? and press RETURN to search for entry-points");

      t2fAq = new JTextField(4);

      t2fAq.setMaximumSize(t2fAq.getPreferredSize());

      
      t2lEpName = new JLabel();
      JPanel pnl = AdminUtils.groupComponents(
            new Component[][] { { t2fEpId, t2lEpName } });
      // remove auto-gapings from the panel
      GroupLayout layout = (GroupLayout) pnl.getLayout();
      layout.setAutoCreateContainerGaps(false);
      layout.setAutoCreateGaps(false);

      t2fEpId.setMaximumSize(t2fEpId.getPreferredSize());

      // build a matrix with the layout
      final Component[][] comps = 
      {
         { new JLabel(" Entry Point "), pnl },
         { new JLabel(" Auto Quit "), t2fAq },
      };

      return AdminUtils.groupComponents(comps, "Menu Defaults");
   }

   /**
    * Build the account-data panel. The following widgets are here: employee,
    * customer and user-code.
    * 
    * @return   a panel with account-data widgets
    */
   private JPanel buildAccountDataPanel()
   {
      t1fEId = new JTextField(8);
      t1fCId = new JTextField(8);

      t1fEId.setToolTipText(
         "Enter ? and press RETURN to search for employees");
      t1fCId.setToolTipText(
         "Enter ? and press RETURN to search for customers");
      
      t1lEName = new JLabel();
      t1lCName = new JLabel();
      
      t1fEId.setMaximumSize(t1fEId.getPreferredSize());
      t1fCId.setMaximumSize(t1fCId.getPreferredSize());
      
      t1chkHe = new JCheckBox();

      JPanel pe = AdminUtils.groupComponents(
            new Component[][] { { t1fEId, t1lEName } });
      // remove auto-gapings from the panel
      GroupLayout layoute = (GroupLayout) pe.getLayout();
      layoute.setAutoCreateContainerGaps(false);
      layoute.setAutoCreateGaps(false);

      JPanel pc = AdminUtils.groupComponents(
            new Component[][] { { t1fCId, t1lCName } });
      // remove auto-gapings from the panel
      GroupLayout layoutc = (GroupLayout) pc.getLayout();
      layoutc.setAutoCreateContainerGaps(false);
      layoutc.setAutoCreateGaps(false);
      
      t1fUc = new JTextField(3);
      t1fUc.setEnabled(false);
      t1fUc.setMaximumSize(t1fUc.getPreferredSize());

      JLabel t1lCd = new JLabel(" Date Created ");
      t1fCd = new JTextField(8);
      t1fCd.setEnabled(false);
      t1fCd.setMaximumSize(t1fCd.getPreferredSize());
      
      // group the user-code and created-date fields
      JPanel puc = AdminUtils.groupComponents(
            new Component[][] { { t1fUc, t1lCd, t1fCd } });
      // remove auto-gapings from the panel
      GroupLayout layoutuc = (GroupLayout) puc.getLayout();
      layoutuc.setAutoCreateContainerGaps(false);
      layoutuc.setAutoCreateGaps(false);
      
      // build a matrix with the layout
      final Component[][] comps = 
      {
         { new JLabel(" MAJIC Configuration "), t1chkHe },
         { new JLabel(" Employee "), pe },
         { new JLabel(" Customer "), pc },
         { new JLabel(" User Code "), puc },
      };

      JPanel pnl = AdminUtils.groupComponents(comps);
      pnl.setBorder(new EtchedBorder(EtchedBorder.RAISED));
      
      return pnl;
   }

   /**
    * Build the panel with default account data.
    * 
    * @return   a panel with default account widgets
    */
   private JPanel buildAccountDefaultsPanel()
   {
      JPanel pnl = ias.getDefaultWidgetPanel();

      pnl.setBorder(new TitledBorder
            (new EtchedBorder(EtchedBorder.RAISED), " Account defaults "));

      return pnl;
   }

   /**
    * Client-side validation of the extension data. This means only following
    * validation is done:
    * <ul>
    *    <li>check if the mandatory widgets have data</li>
    *    <li>check if the numeric, date, etc fields are valid</li>
    *    <li>any other special validation which can be done on client-side</li>
    * </ul>
    * 
    * @param    accountName
    *           The account for which extension field validation is needed. If 
    *           <code>null</code>, it means a new account is being added.
    * 
    * @return   the extension object which holds extension data. may be null
    *           if extensions are not required.
    * 
    * @throws   AccountExtValidationException if client-side validation fails
    */
   public Serializable validateExtensions(String accountName)
   throws AccountExtValidationException
   {
      // first step: check if it uses extensions
      boolean hasExt = t1chkHe.isSelected();
      
      // if no extension data is needed, bypass validation and return null
      if (!hasExt)
      {
         return null;
      }
      
      // validate integer fields; no field is mandatory
      int empNo = validateIntegerField(t1fEId, "invalid employee number");
      int custNo = validateIntegerField(t1fCId, "invalid customer number");
      int autoQuit = validateIntegerField(t2fAq, "invalid auto-quit value");
      int userLevel = validateIntegerField(t4fUl, "invalid user-level value");

      // employee
      extDef.empNum = empNo;

      // customer
      extDef.custNum = custNo;
      
      // user-code
      extDef.userCode = t1fUc.getText().trim(); // rechecked on server-side
         
      // entry point
      extDef.entryPoint = t2fEpId.getText().trim();
         
      // auto-quit
      extDef.autoQuit = autoQuit;
         
      // user-help
      extDef.userHelp = t2fUh.getText().trim();
         
      // run-before
      extDef.runBefore = t2chkRb.isSelected();
         
      // printer
      extDef.printer = t3fPrId.getText().trim();
         
      // default file
      extDef.defaultFile = t3fDf.getText().trim();
         
      // file viewer
      extDef.fileViewer = t3fFv.getText().trim();
         
      // file editor
      extDef.fileEditor = t3fFe.getText().trim();
         
      // top dir
      extDef.topDir = t3fTd.getText().trim();
         
      // browse mask
      extDef.browseMask = t4chkBm.isSelected();
         
      // user-level
      extDef.userLevel = userLevel;
         
      // valid apps
      extDef.validApps = t4fVa.getText().trim();
         
      // description
      extDef.description = t5fDesc.getText().trim();
      
      if (extDef.description.length() > 40)
      {
         throw new AccountExtValidationException("description must be " +
         		" max 40 characters long");
      }

      return extDef;
   }
      
   /**
    * Validate the specified field to be an integer value. If the value is not
    * an integer field, then an {@link AccountExtValidationException} is throw
    * with the specified message.
    * 
    * @param    tf
    *           The field to be validated.
    * @param    msg
    *           The message to be displayed is value is not valid.
    * 
    * @return   the integer value of this field
    * 
    * @throws   AccountExtValidationException
    *           if the field doesn't contain a valid integer value.
    */
   private int validateIntegerField(JTextField tf, String msg)
   throws AccountExtValidationException 
   {
      String txt = tf.getText().trim();
      try
      {
         return Integer.parseInt(txt);
      }
      catch (Exception e) 
      {
         throw new AccountExtValidationException(msg);
      }
   }

   /**
    * As this implementation uses all-widget-layout, it returns null;
    * 
    * @return   <code>null</null>
    */
   public Container getExtContainer()
   {
      return null;
   }
   
   /**
    * This is a no-op.
    * 
    * @param    c
    *           the container to which widgets will be added.
    * @param    accountName
    *           The account for which extension field layout is needed. If 
    *           <code>null</code>, it means a new account is added.
    * 
    * @param    writable
    *           <code>true</code> if this is for enabled inputs.
    *           <code>false</code> for read only panel
    */
   public void layoutExtWidgets(Container c, String accountName, boolean writable)
   {
      // no-op
   }

   /**
    * Save the current {@link IAccountScreen} instance to be used in the 
    * plugin.
    * 
    * @param   screen
    *          the current {@link IAccountScreen} instance.
    */
   public void initialize(IAccountScreen screen)
   {
      this.ias = screen;
   }

   /**
    * Returns the print preview table headers for the account extension data.
    * An empty array disables extension data printout.
    * 
    * @return   Array of table headers.
    */
   public String[] getPreviewHeaders()
   {
      return new String[] {"Employee",       // 0
                           "Customer",       // 1
                           "User Code",      // 2
                           "Date Created",   // 3
                           "Entry Point",    // 4
                           "Auto Quit",      // 5
                           "User Help",      // 6
                           "Run Before",     // 7
                           "Printer",        // 8
                           "Default File",   // 9
                           "File Viewer",    // 10
                           "File Editor",    // 11
                           "Top Directory",  // 12
                           "Browse Mask",    // 13
                           "User Level",     // 14
                           "Valid Apps",     // 15
                           "Description"     // 16
                           };
   }
   
   /**
    * Gathers and returns all extra columns info for the specific account as
    * strings. The number and the order of values should correspond to the
    * headers returned by {@link #getPreviewHeaders}.
    * 
    * @param    subjectId
    *           account name
    *           
    * @return   Array of extra column info strings. 
    */
   public String[] getPreviewRow(String subjectId)
   {
      // the results are here
      String[] ext = new String[] {"",
                                   "",
                                   "",
                                   "",
                                   "",
                                   "",
                                   "",
                                   "",
                                   "",
                                   "",
                                   "",
                                   "",
                                   "",
                                   "",
                                   "",
                                   "",
                                   ""
                                  };
                           
      // get extension data
      MajicExtDef extDef = majic.getAccountExtensions(subjectId);
      
      if (extDef == null)
      {
         return ext;
      }
      
      // employee
      ext[0] = String.format("%d", extDef.empNum);
      
      // customer
      ext[1] = String.format("%d", extDef.custNum);
      
      // user-code
      ext[2] = extDef.userCode != null ? extDef.userCode : "?";
      
      // date created
      ext[3] = extDef.createdDate != null ? extDef.createdDate.toString() : "?";
      
      // entry point
      ext[4] = extDef.entryPoint != null ? extDef.entryPoint : "?";
      
      // auto-quit
      ext[5] = String.format("%d", extDef.autoQuit);
      
      // user-help
      ext[6] = extDef.userHelp != null ? extDef.userHelp : "?";
      
      // run-before
      ext[7] = extDef.runBefore ? "yes" : "no";
      
      // printer
      ext[8] = extDef.printer != null ? extDef.printer : "?";
      
      // default file
      ext[9] = extDef.defaultFile != null ? extDef.defaultFile : "?";
      
      // file viewer
      ext[10] = extDef.fileViewer != null ? extDef.fileViewer : "?";
      
      // file editor
      ext[11] = extDef.fileEditor != null ? extDef.fileEditor : "?";
      
      // top dir
      ext[12] = extDef.topDir != null ? extDef.topDir : "?";
      
      // browse mask
      ext[13] = extDef.browseMask ? "yes" : "no";
      
      // user-level
      ext[14] = String.format("%d", extDef.userLevel);
      
      // valid apps
      ext[15] = extDef.validApps != null ? extDef.validApps : "?";
      
      // description
      ext[16] = extDef.description != null ? extDef.description : "?";

      return ext;
   }
   
   /**
    * Returns the extended print preview table headers for the account
    * extension data. An empty array disables extended print for the extension
    * data printout.
    * 
    * @return   Array of headers.
    */
   public String[] getExtendedHeaders()
   {
      return new String[] {
                             "Attribute Name",
                             "Attribute Value",
                             "Attribute Name",
                             "Attribute Value",
                          }; 
   }
   
   /**
    * Returns the extended print preview table rows required for the account
    * extension data per account.
    * 
    * @return   Number of extended rows.
    */
   public int getExtendedRows()
   {
      // 2 Account
      //    Employee   xxxxxxxxxx     User Code     xxxxx
      //    Customer   xxxxxxxxxx     Date Created  xxxxxxxxx
      // 2 Application Defaults
      //    Entry Point   xxxxxxxxxx  User Help     xxxxxxxxxxxx
      //    Auto Quit     xxxxxxxxxx  Run Before    xxx
      // 3 System Defaults
      //    Printer       xxxxxxxxxx   
      //    File Viewer   xxxxxxxxxx  Default File  xxxxxxxxxxxxx 
      //    File Editor   xxxxxxxxxx  Top Directory xxxxxxxxxxxxx
      // 2 Security
      //    Browse Mask   xxx         User level    xxxxxxxx 
      //    Valid Apps    xxxxxxxxxx 
      // 1 Description
      //    Description   xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
      //
      return EXTENDED_ROWS;
   }
   
   /**
    * Fills the extended print preview table row columns with the account
    * extension data.
    * 
    * @param    subjectId
    *           account name
    * 
    * @param    tm
    *           table model to be used
    * 
    * @param    row
    *           index of the table model row to start filling from
    * 
    * @param    col
    *           index of the table column to start filling from
    */
   public void fillExtendedRow(String subjectId, TableModel tm, int row,
                               int col)
   {
      // get extension data
      MajicExtDef extDef = majic.getAccountExtensions(subjectId);
      
      if (extDef == null)
      {
         // no extension data; leave all blanks
         for (int i = 0; i < EXTENDED_ROWS; i++)
         {
            for (int j = 0; j < EXTENDED_COLS; j++)
            {
               tm.setValueAt("", row + i, col + j);
            }
         }
         
         return;
      }
      
      String v = null;
      
      // employee
      v = String.format("%d", extDef.empNum);
      tm.setValueAt("Employee",                            row + 0, col + 0);
      tm.setValueAt(v,                                     row + 0, col + 1);
      
      // user-code
      v = extDef.userCode != null ? extDef.userCode : "?";
      tm.setValueAt("User Code",                           row + 0, col + 2);
      tm.setValueAt(v,                                     row + 0, col + 3);
      
      // customer
      v = String.format("%d", extDef.custNum);
      tm.setValueAt("Customer",                            row + 1, col + 0);
      tm.setValueAt(v,                                     row + 1, col + 1);
      
      // date created
      v = extDef.createdDate != null ? extDef.createdDate.toString() : "?";
      tm.setValueAt("Date Created",                        row + 1, col + 2);
      tm.setValueAt(v,                                     row + 1, col + 3);
      
      // entry point
      v = extDef.entryPoint != null ? extDef.entryPoint : "?";
      tm.setValueAt("Entry Point",                         row + 2, col + 0);
      tm.setValueAt(v,                                     row + 2, col + 1);
      
      // user-help
      v = extDef.userHelp != null ? extDef.userHelp : "?";
      tm.setValueAt("User Help",                           row + 2, col + 2);
      tm.setValueAt(v,                                     row + 2, col + 3);
      
      // auto-quit
      v = String.format("%d", extDef.autoQuit);
      tm.setValueAt("Auto Quit",                           row + 3, col + 0);
      tm.setValueAt(v,                                     row + 3, col + 1);
      
      // run-before
      v = extDef.runBefore ? "yes" : "no";
      tm.setValueAt("Run Before",                          row + 3, col + 2);
      tm.setValueAt(v,                                     row + 3, col + 3);
      
      // printer
      v = extDef.printer != null ? extDef.printer : "?";
      tm.setValueAt("Printer",                             row + 4, col + 0);
      tm.setValueAt(v,                                     row + 4, col + 1);
      
      // blank
      tm.setValueAt("",                                    row + 4, col + 2);
      tm.setValueAt("",                                    row + 4, col + 3);
      
      // file viewer
      v = extDef.fileViewer != null ? extDef.fileViewer : "?";
      tm.setValueAt("File Viewer",                         row + 5, col + 0);
      tm.setValueAt(v,                                     row + 5, col + 1);
      
      // default file
      v = extDef.defaultFile != null ? extDef.defaultFile : "?";
      tm.setValueAt("Default File",                        row + 5, col + 2);
      tm.setValueAt(v,                                     row + 5, col + 3);
      
      // file editor
      v = extDef.fileEditor != null ? extDef.fileEditor : "?";
      tm.setValueAt("File Editor",                         row + 6, col + 0);
      tm.setValueAt(v,                                     row + 6, col + 1);
      
      // top dir
      v = extDef.topDir != null ? extDef.topDir : "?";
      tm.setValueAt("Top Directory",                       row + 6, col + 2);
      tm.setValueAt(v,                                     row + 6, col + 3);
      
      // browse mask
      v = extDef.browseMask ? "yes" : "no";
      tm.setValueAt("Browse Mask",                         row + 7, col + 0);
      tm.setValueAt(v,                                     row + 7, col + 1);
      
      // user-level
      v = String.format("%d", extDef.userLevel);
      tm.setValueAt("User Level",                          row + 7, col + 2);
      tm.setValueAt(v,                                     row + 7, col + 3);
      
      // valid apps
      v = extDef.validApps != null ? extDef.validApps : "?";
      tm.setValueAt("Valid Apps",                          row + 8, col + 0);
      tm.setValueAt(v,                                     row + 8, col + 1);
      
      // blank
      tm.setValueAt("",                                    row + 8, col + 2);
      tm.setValueAt("",                                    row + 8, col + 3);
      
      // description
      v = extDef.description != null ? extDef.description : "?";
      tm.setValueAt("Description",                         row + 9, col + 0);
      tm.setValueAt(v,                                     row + 9, col + 1);
      
      // blank
      tm.setValueAt("",                                    row + 9, col + 2);
      tm.setValueAt("",                                    row + 9, col + 3);
   }

   /**
    * Create and return the panel containing user search elements.
    *
    * @param   findScreen
    *          Find screen which handles user searching.
    *
    * @return  See above.
    */
   public JPanel buildUserFindPanel(final IFindAccountScreen findScreen)
   {
      findUserInput = new JTextField();
      findUserInput.setToolTipText("You can press ENTER to perform search.");

      JPanel optionsPanel = new JPanel();
      optionsPanel.setLayout(new BoxLayout(optionsPanel, BoxLayout.Y_AXIS));
      optionsPanel.add(new JLabel("Enter User Name or Employee Number:"));
      optionsPanel.add(Box.createRigidArea(new Dimension(0,5)));
      optionsPanel.add(findUserInput);

      // add Find and Cancel buttons
      JPanel buttonsPanel = new JPanel();
      buttonsPanel.add(findScreen.getCancelButton());
      buttonsPanel.add(findScreen.getFindButton());

      JPanel targetPanel = AdminUtils.groupComponents(new Component[][]
      {
         { optionsPanel },
         { buttonsPanel }
      });

      findUserInput.addKeyListener(new KeyListener()
      {
         public void keyTyped(KeyEvent keyEvent)
         {
         }

         public void keyPressed(KeyEvent keyEvent)
         {
            if (keyEvent.getKeyCode() == KeyEvent.VK_ENTER)
            {
               handlePerformUserFind(findScreen);
            }
         }

         public void keyReleased(KeyEvent keyEvent)
         {
         }
      });

      // create event handlers: find window
      findScreen.getFindButton().addActionListener(new ActionListener()
      {
         @Override
         @SuppressWarnings("unused")
         public void actionPerformed(ActionEvent evt)
         {
            handlePerformUserFind(findScreen);
         }
      });

      findScreen.getCancelButton().addActionListener(new ActionListener()
      {
         @Override
         @SuppressWarnings("unused")
         public void actionPerformed(ActionEvent evt)
         {
            findScreen.handleCancel();
         }
      });

      return targetPanel;
   }

   /**
    * Attaches the plugin to the AdminClient instance. Called once upon plugin
    * loading.
    * 
    * @param    client
    *           instance of AdminClient
    */
   public void attach(AdminClient client)
   {
      this.client = client;
      
      // reset aged data
      majic = null;
      majicMenu = null;
      fun = null;
      psr = null;
      psrCount = 0;
      comp = null;
      
      // get the majic exports
      majic = (MajicExtExports) 
              RemoteObject.obtainNetworkInstance(MajicExtExports.class, 
                                                 client.getSession());
      
      client.add(buildPrintMaintenance()   , SCREEN_PRINTER_MAINTENANCE, new PrinterPrinter());
      client.add(buildFunctionMaintenance(), SCREEN_FUNCTION_MAINTENANCE, new FunctionPrinter());
      client.add(buildPanelMenuBuilder()   , SCREEN_MENU_BUILDER, new MenuPrinter());
   }

   /**
    * Populate printers data model with data.
    */
   private void populatePrinters()
   {
      printers.clearSelection();
      printerTm.clear();
      
      // get the list of printers from the server
      AdminUtils.busy(client);
      TaggedName[] list = majic.listAllPrinters();
      AdminUtils.idle(client);
      
      if (list != null)
      {
         if (list.length > 0)
         {
            printerTm.loadAll(list);
         }
      }
      else
      {
         client.alarm();
      }
   }

   /**
    * Populate system functions model with data.
    */
   private void populateFunctions()
   {
      functions.clearSelection();
      functionTm.clear();
      funcCache = null;
      
      // get the list of functions from the server
      AdminUtils.busy(client);
      TaggedName[] list = majic.listAllFunctions();
      AdminUtils.idle(client);
      
      if (list != null)
      {
         if (list.length > 0)
         {
            // fill all functions
            
            Map<String, TaggedName> map = new HashMap<String, TaggedName>();
            
            for (TaggedName name : list)
            {
               map.put(name.get(2), name);
            }

            Map<String, Integer> acls = majic.getACLs();
            
            // check if function has related ACLs
            for (String name : map.keySet())
            {
               if (!acls.containsKey(name))
                  continue;
               
               TaggedName tname = map.get(name);
               
               // appended status "A" means that function has related ACLs
               StringBuilder sb = new StringBuilder(tname.get(0));
               sb.append("A");
               tname.set(0, sb.toString());
            }
            
            TaggedName[] values = map.values().toArray(new TaggedName[map.size()]);
            
            Arrays.sort(values, new Comparator<TaggedName>() 
                        {
                           @Override
                           public int compare(TaggedName o1, TaggedName o2)
                           {
                              return o1.get(2).compareTo(o2.get(2));
                           }
               
                        });
            
            functionTm.loadAll(values);
         }
         
         majic.refreshFunctions();
      }
      else
      {
         client.alarm();
      }
   }

   /**
    * Populate panel menus model with data.
    */
   private void populateMenus()
   {
      menuMap.clear();
      menus.clearSelection();
      menuListTm.clear();
      
      // get the list of menus from the server
      AdminUtils.busy(client);
      MenuDef[] list = majic.listMenus(null);
      AdminUtils.idle(client);
      
      if (list != null)
      {
         for (MenuDef menuDef : list)
         {
            menuMap.put(menuDef.getName(), menuDef);
         }
         
         TaggedName[] menus = new TaggedName[menuMap.size()];
         
         int i = 0;
         
         for (String name : menuMap.keySet())
         {
            menus[i++] = menuMap.get(name).getMenu();
         }
         
         if (list.length > 0)
         {
            menuListTm.loadAll(menus);
         }
      }
      else
      {
         client.alarm();
      }
   }
   
   /**
    * Populate panel menu items model with data.
    */
   private void populateMenuItems()
   {
      menuItems.clearSelection();
      menuItemsTm.clear();
      
      int row = menus.getSelectedRow();
      
      if (row < 0)
         return;
      
      MenuDef def = menuMap.get(menuListTm.getValueAt(row, 0));
      
      if (def == null)
         return;
      
      if (!def.isComplete())
      {
         AdminUtils.busy(client);
         MenuDef[] list = majic.listMenus((String) menuListTm.getValueAt(row, 0));
         AdminUtils.idle(client);
         
         if (list != null && list.length == 1)
         {
            menuMap.put(list[0].getName(), list[0]);
            def = list[0];
         }
         else
         {
            client.beep("Unable to retrieve menu items");
            return;
         }
      }
      
      TaggedName[] list = def.getMenuItemsNarrow(); 
      
      if (list != null && list.length > 0)
      {
         menuItemsTm.loadAll(list);
      }
      else
      {
         client.alarm();
      }
   }
   
   /**
    * Prepare printer maintenance screen.
    * 
    * @return   Prepared panel.
    */
   private JPanel buildPrintMaintenance()
   {
      JPanel panel = new JPanel();
      
      panel.setLayout(new BorderLayout());
      
      printerTm = new TaggedDataTableModel("Name",
                                           "Type",
                                           "Input File",
                                           "OutputFile",
                                           "Start String",
                                           "End String",
                                           "Star Procedure",
                                           "End Procedure");
      printers = new JTable(printerTm);
      printers.getTableHeader().setReorderingAllowed(false);
      
      JScrollPane scroller = new JScrollPane();
      scroller.getViewport().add(printers);
      scroller.setBorder(new TitledBorder
         (new EtchedBorder(EtchedBorder.RAISED), " Defined Printers "));
      
      panel.add(scroller, BorderLayout.CENTER);

      JButton refresh      = new JButton("Refresh");
      JButton add          = new JButton("Add Printer");
      final JButton edit   = new JButton("Edit Printer");
      final JButton delete = new JButton("Delete Selected");

      JPanel wpCtlPanel = AdminUtils.groupComponents(new Component[][] 
         { 
            { refresh, add, edit, new JLabel(), new JLabel(), delete }
         },
         true
      );
      panel.add(wpCtlPanel, BorderLayout.SOUTH);
      
      toggleButtons(false, edit, delete);
      
      printers.getSelectionModel().addListSelectionListener(new ListSelectionListener()
      {
         @Override
         @SuppressWarnings("unused") 
         public void valueChanged(ListSelectionEvent e)
         {
            toggleButtons(printers.getSelectedRowCount() > 0, edit, delete);
         }
      });
      
      add.addActionListener(new ActionListener()
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent e)
         {
            // add
            addOrUpdatePrinter();
         }
      });

      edit.addActionListener(new ActionListener()
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent e)
         {
            if (printers.getSelectedRowCount() == 0)
            {
               // should not happen
               client.beep("no selected row");
               return;
            }
            
            addOrUpdatePrinter(getRowValues(printerTm, printers.getSelectedRow(), 0));
         }
      });

      delete.addActionListener(new ActionListener()
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent e)
         {
            if (printers.getSelectedRowCount() == 0)
            {
               // should not happen
               client.beep("no selected row");
               return;
            }
            
            int[] rows = printers.getSelectedRows();

            Boolean res = AdminUtils.confirmCancelDialog(client, 
                                                         makePlural(rows.length, 
                                                                    "Delete record"), 
                                                         "You are about to delete " 
                                                         + rows.length 
                                                         + makePlural(rows.length, 
                                                                      " printer definition")
                                                         + ".\nAre you sure?");

            if (!res.booleanValue())
               return;
            
            AdminUtils.busy(client);
            
            boolean cleanup = true;
            
            for (int row : rows)
            {
               String[] values = getRowValues(printerTm, row, 0);
               
               if (values == null)
               {
                  // should not happen
                  client.beep("no selected row");
                  continue;
               }

               int rc = majic.printerDbHelper(new TaggedName(values), 
                                              MajicExtExports.MODE_DELETE);
               
               if (rc != 0)
               {
                  cleanup = false;
                  client.beep("ERROR: unable to delete printer " + values[0]);
               }
            }
            
            if (cleanup)
               client.ok();
            
            populatePrinters();
            
            AdminUtils.idle(client);
         }
      });
      
      refresh.addActionListener(new ActionListener()
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent e)
         {
            // refresh
            populatePrinters();
         }
      });
      
      return panel;
   }
   
   /**
    * Prepare function maintenance screen.
    * 
    * @return   Prepared panel.
    */
   private JPanel buildFunctionMaintenance()
   {
      JPanel panel = new JPanel();
      
      panel.setLayout(new BorderLayout());
      
      functionTm = new TaggedDataTableModel(// special marks
                                            "State",
                                            // regular fields
                                           "Short Name",
                                           "Full Name",
                                           "Type",
                                           "ID",
                                           "Keystroke Access",
                                           "Printer Name",
                                           "Program To Run",
                                           "Modifycation Date",
                                           "Transaction",
                                           "Symix Function",
                                           "Priority",
                                           "Run/Call",
                                           "Permit",
                                           "Type Decription",
                                           "Approval Code",
                                           "Resp. Dept.");
      functionTm.setSort(false);
      
      functions = new JTable(functionTm);
      functions.getTableHeader().setReorderingAllowed(false);
      
      int width[] = {10, 60, 100, 50, 10, 25, 20, 60, 
         40, 30, 30, 30, 30, 30, 50, 60, 60};

      setColumnWidth(functions, width);
      
      JScrollPane scroller = new JScrollPane();
      scroller.getViewport().add(functions);
      scroller.setBorder(new TitledBorder
         (new EtchedBorder(EtchedBorder.RAISED), " Defined Functions "));
      
      panel.add(scroller, BorderLayout.CENTER);

      JButton refresh      = new JButton("Refresh");
      JButton add          = new JButton("Add Function");
      final JButton edit   = new JButton("Edit Function");
      final JButton delete = new JButton("Delete Selected");
      final JButton acl    = new JButton("Edit ACL");

      JPanel wpCtlPanel = AdminUtils.groupComponents(new Component[][] 
         { 
            { refresh, add, edit, acl, new JLabel(), new JLabel(), delete }
         },
         true
      );
      panel.add(wpCtlPanel, BorderLayout.SOUTH);
      
      final JButton[] selList = new JButton[] {delete, edit, acl};
      
      toggleButtons(false, selList);
      
      functions.getSelectionModel().addListSelectionListener(new ListSelectionListener()
      {
         @Override
         @SuppressWarnings("unused") 
         public void valueChanged(ListSelectionEvent e)
         {
            toggleButtons(functions.getSelectedRowCount() > 0, selList);
         }
      });
      
      add.addActionListener(new ActionListener()
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent e)
         {
            addOrUpdateFunction();
         }
      });

      edit.addActionListener(new ActionListener()
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent e)
         {
            if (functions.getSelectedRowCount() == 0)
            {
               // should not happen
               client.beep("no selected row");
               return;
            }
            
            addOrUpdateFunction(getRowValues(functionTm, functions.getSelectedRow(), 1));
         }
      });

      delete.addActionListener(new ActionListener()
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent e)
         {
            if (functions.getSelectedRowCount() == 0)
            {
               // should not happen
               client.beep("no selected row");
               return;
            }
            
            int[] rows = functions.getSelectedRows();

            Boolean res = AdminUtils.confirmCancelDialog(client, 
                                                         makePlural(rows.length, 
                                                                    "Delete record"), 
                                                         "You are about to delete " 
                                                         + rows.length 
                                                         + makePlural(rows.length, 
                                                                      " function definition")
                                                         + ".\nAre you sure?");

            if (!res.booleanValue())
               return;
            
            AdminUtils.busy(client);
            
            boolean cleanup = true;
            
            for (int row : rows)
            {
               String[] values = getRowValues(functionTm, row, 1);
               
               if (values == null)
               {
                  // should not happen
                  client.beep("no selected row");
                  continue;
               }

               List<String> valuesList = new ArrayList<String>(Arrays.asList(values));
               valuesList.add("");
               valuesList.add("");
               valuesList.add("");
               valuesList.add("");
               valuesList.add("");
               valuesList.add("");
               valuesList.add(valuesList.get(1)); // "old function name"

               int rc = majic.functionDbHelper(new TaggedName(valuesList.toArray(new String[0])),
                                               MajicExtExports.MODE_DELETE);
               
               if (rc != 0)
               {
                  cleanup = false;
                  client.beep("ERROR: unable to delete function " + values[0]);
               }
            }
            
            if (cleanup)
               client.ok();
            
            populateFunctions();
            
            AdminUtils.idle(client);
         }
      });
      
      refresh.addActionListener(new ActionListener()
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent e)
         {
            // refresh
            populateFunctions();
         }
      });
      
      acl.addActionListener(new ActionListener()
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent e)
         {
            if (functions.getSelectedRowCount() == 0)
            {
               // should not happen
               client.beep("no selected row");
               return;
            }
            
            editFunctionACL(getRowValues(functionTm, functions.getSelectedRow(), 1)[1]);
         }
      });

      return panel;
   }

   /**
    * Set width of columns in table.
    * 
    * @param    table
    * @param    width
    */
   private void setColumnWidth(JTable table, int[] width)
   {
      for (int i = 0; i < table.getColumnCount(); i++)
      {
         table.getColumnModel().getColumn(i).setPreferredWidth(width[i]);
      }
   }

   /**
    * Edit function ACL
    * 
    * @param    name
    */
   private void editFunctionACL(String name)
   {
      AdminACL aacl = new AdminACL(client, true, name);
      
      if (!majic.addFunction(name))
         return;
      
      final JDialog dialog = AdminUtils.modalDialog(client, 
                                                    "Function ACL", 
                                                    aacl.getMainPanel());
      
      aacl.setDialog(dialog);
      
      dialog.setVisible(true);
      populateFunctions();
   }

   /**
    * Toggle state of listed buttons.
    * @param    b
    *           <code>true</code> - enable buttons, <code>false</code> -
    *           disable buttons.
    * @param    selList
    *           List of buttons.
    */
   static protected void toggleButtons(boolean b, JButton... selList)
   {
      for (JButton button : selList)
      {
         button.setEnabled(b);
      }
   }
   
   /**
    * Prepare menu builder screen.
    * 
    * @return   Prepared panel.
    */
   private JPanel buildPanelMenuBuilder()
   {
      JPanel panel = new JPanel();
      
      panel.setLayout(new BorderLayout());
      
      menuListTm = new TaggedDataTableModel("Name",
                                            "Description",
                                            "Narrow");
      menuItemsTm = new TaggedDataTableModel("Key",
                                             "Description",
                                             "Function");
      
      menuItemsTm.setSort(false);
      
      menus = new JTable(menuListTm);
      menus.getTableHeader().setReorderingAllowed(false);
      
      setColumnWidth(menus, new int[] {50, 200, 10});
      
      JScrollPane scroller = new JScrollPane();
      scroller.getViewport().add(menus);
      scroller.setBorder(new TitledBorder
         (new EtchedBorder(EtchedBorder.RAISED), " Defined Panel Menus "));
      
      panel.add(scroller, BorderLayout.CENTER);

      menuItems = new JTable(menuItemsTm);
      menuItems.getTableHeader().setReorderingAllowed(false);

      setColumnWidth(menuItems, new int[] {10, 200, 90, 80});
      
      scroller = new JScrollPane();
      scroller.getViewport().add(menuItems);
      scroller.setBorder(new TitledBorder
         (new EtchedBorder(EtchedBorder.RAISED), " Defined Menu Items"));
      
      panel.add(scroller, BorderLayout.EAST);
      
      JButton refresh      = new JButton("Refresh");
      JButton add          = new JButton("Add Menu");
      final JButton edit   = new JButton("Edit Menu");
      final JButton delete = new JButton("Delete Selected");

      JPanel wpCtlPanel = AdminUtils.groupComponents(new Component[][] 
         { 
            { refresh, add, edit, new JLabel(), new JLabel(), delete}
         },
         true
      );
      panel.add(wpCtlPanel, BorderLayout.SOUTH);
      
      toggleButtons(false, delete, edit);
      
      menus.getSelectionModel().addListSelectionListener(new ListSelectionListener()
      {
         @Override
         @SuppressWarnings("unused") 
         public void valueChanged(ListSelectionEvent e)
         {
            toggleButtons(menus.getSelectedRowCount() > 0, delete, edit);
            
            if (menus.getSelectedRowCount() == 1)
            {
               populateMenuItems();
            }
            else
            {
               menuItems.clearSelection();
               menuItemsTm.clear();
            }
         }
      });
      
      add.addActionListener(new ActionListener()
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent e)
         {
            // add
            addOrUpdateMenu();
         }
      });

      edit.addActionListener(new ActionListener()
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent e)
         {
            if (menus.getSelectedRowCount() == 0)
            {
               // should not happen
               client.beep("no selected row");
               return;
            }
            
            addOrUpdateMenu(getRowValues(menuListTm, menus.getSelectedRow(), 0));
         }
      });

      delete.addActionListener(new ActionListener()
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent e)
         {
            if (menus.getSelectedRowCount() == 0)
            {
               // should not happen
               client.beep("no selected row");
               return;
            }
            
            int[] rows = menus.getSelectedRows();

            Boolean res = AdminUtils.confirmCancelDialog(client, 
                                                         makePlural(rows.length, 
                                                                    "Delete record"), 
                                                         "You are about to delete " 
                                                         + rows.length 
                                                         + makePlural(rows.length, 
                                                                      " menu definition")
                                                         + ".\nAre you sure?");

            if (!res.booleanValue())
               return;
            
            AdminUtils.busy(client);
            
            boolean cleanup = true;
            MenuDef def = new MenuDef();
            
            for (int row : rows)
            {
               String[] values = getRowValues(menuListTm, row, 0);
               
               if (values == null)
               {
                  // should not happen
                  client.beep("no selected row");
                  continue;
               }
               
               
               def.setName(values[0]);

               try
               {
                  int rc = majic.menuDbHelper(def, MajicExtExports.MODE_DELETE);

                  if (rc != 0)
                  {
                     cleanup = false;
                     client.beep("ERROR: unable to delete menu " + values[0]);
                     break;
                  }
               }
               catch (Exception v)
               {
                  cleanup = false;
                  client.beep("ERROR: menu " + values[0] +
                              " can't be deleted, " + v.getMessage());
                  break;
               }
            }
            
            if (cleanup)
               client.ok();
            
            populateMenus();
            
            AdminUtils.idle(client);
         }
      });
      
      refresh.addActionListener(new ActionListener()
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent e)
         {
            // refresh
            populateMenus();
         }
      });

      return panel;
   }
   
   /**
    * @param values 
    * 
    */
   protected void addOrUpdateMenu(String ... values)
   {
      final boolean isEdit = values != null && values.length > 0;
      JPanel pnlInput  = new JPanel();
      
      GroupLayout layout = new GroupLayout(pnlInput);
      pnlInput.setLayout(layout);
      layout.setAutoCreateGaps(true);
      layout.setAutoCreateContainerGaps(true);
      
      JLabel lmName   = new JLabel("Menu Name");
      JLabel lmDesc   = new JLabel("Menu Description");

      final JTextField mName   = new JLimitedTextField(14);
      final JTextField mDesc   = new JLimitedTextField(34);
      
      final JLabel[] lbl = new JLabel[] { lmName, lmDesc}; 

      final JTextField[] txt = new JTextField[] { mName, mDesc}; 
      
      setMaximumSize(txt);
      
      mName.setEnabled(!isEdit);

      addInputs(pnlInput, txt);

      GroupLayout.SequentialGroup hGroup = layout.createSequentialGroup();
      
      // parallel group of labels
      addParallelGroup(hGroup, layout, lbl);
      // parallel group of input fields
      addParallelGroup(hGroup, layout, txt);
      
      layout.setHorizontalGroup(hGroup);
      layout.setVerticalGroup(newSequentialGroup(layout, lbl, txt));
      
      layout.linkSize(lbl);

      JPanel pnlItems = new JPanel();
      
      final TaggedDataTableModel menuItemsTmEdit;
      menuItemsTmEdit = new TaggedDataTableModel("Key",
                                                 "Item Text",
                                                 "Function", 
                                                 "Key", 
                                                 "Item Text",
                                                 "Function")
      {
         /**
          * Check if cell can be edited. In this implementation all cells are
          * editable.
          * 
          * @param    rowIndex
          * @param    columnIndex
          */
         @Override
         public boolean isCellEditable(int rowIndex, int columnIndex)
         {
            return true;
         }

         /**
          * Save value after edition.
          * 
          * @see   javax.swing.table.AbstractTableModel#setValueAt(java.lang.Object,
          *        int, int)
          */
         @Override
         public void setValueAt(Object aValue, int rowIndex, int columnIndex)
         {
            data.get(rowIndex).set(columnIndex, (String) aValue);
         }
      };
      
      menuItemsTmEdit.setSort(false);

      TaggedName[] list = new TaggedName[16];
      boolean filled = false;
      
      if (isEdit && values != null && values.length > 1)
      {
         string2JComponent(values, txt);
         
         MenuDef def = menuMap.get(values[0]);
         
         if (def != null)
         {
            filled = true;
            list = def.getMenuItemsWide();
         }
      }
      
      if (!filled)
      {
         for (int i = 0; i < 16; i++)
         {
            list[i] = new TaggedName("", "", "", "", "", "");
         }
      }
      
      menuItemsTmEdit.loadAll(list);
      
      final JDialog[] owner = new JDialog[] {null};

      final JTable menuItemsEdit = new JTable(menuItemsTmEdit) 
      {
         /**
          * @see javax.swing.JTable#editingStopped(javax.swing.event.ChangeEvent)
          */
         @Override
         public void editingStopped(ChangeEvent e)
         {
            int row = getEditingRow();
            int col = getEditingColumn();
            
            super.editingStopped(e);
            
            int ecol = (col == 2) ? 1 : (col == 5) ? 4 : -1;
            
            if (ecol > 0 && ((String)getValueAt(row, ecol)).trim().equals(""))
            {
               String val = (String)getValueAt(row, col);
               val = val.trim();
               
               for (int i = 0; i < functionTm.getRowCount(); i++)
               {
                  String key = ((String) functionTm.getValueAt(i, 1)).trim();
                  
                  if (key.equals(""))
                     continue;
                  
                  if (key.equals(val))
                  {
                     setValueAt(functionTm.getValueAt(i, 2), row, ecol);
                     repaint(getCellRect(row, ecol, false));
                     return;
                  }
               }
            }
         }
      };
      
      menuItemsEdit.getSelectionModel().setSelectionInterval(0, 0);
      menuItemsEdit.getTableHeader().setReorderingAllowed(false);
      
      int[] width = new int[] {30, 400, 200, 30, 400, 200};
      
      setColumnWidth(menuItemsEdit, width);
      
      menuItemsEdit.setMinimumSize(new Dimension(sum(width), 
                                                 menuItemsEdit.getMinimumSize().height));
      
      menuItemsEdit.getColumnModel().getColumn(2).setCellEditor(new ComboBoxEditor(extractFuncList()));
      menuItemsEdit.getColumnModel().getColumn(5).setCellEditor(new ComboBoxEditor(extractFuncList()));      
      
      pnlItems.setLayout(new BorderLayout());
      
      JScrollPane scroller = new JScrollPane();
      scroller.getViewport().add(menuItemsEdit);
      scroller.setBorder(new TitledBorder
         (new EtchedBorder(EtchedBorder.RAISED), " Menu Items"));
      
      pnlItems.add(scroller, BorderLayout.CENTER);
      
      JPanel pnl2 = null;

      JButton cancel = new JButton("Cancel");
      JButton save   = new JButton("Save");
      
      pnl2 = AdminUtils.groupComponents(new Component[][] {
            { cancel, save },
      });
      layout = (GroupLayout) pnl2.getLayout();
      layout.setAutoCreateContainerGaps(false);
      
      pnlInput.setBorder(new TitledBorder
                         (new EtchedBorder(EtchedBorder.RAISED), " Menu Definition "));
      
      JPanel groupComponents = AdminUtils.groupComponents(new Component[][] 
         { { pnlInput }, { pnlItems }, { pnl2 } } );
      
      final JDialog dialog = AdminUtils.modalDialog(client, 
                                                    "Menu Definition", 
                                                    groupComponents);
      
      owner[0] = dialog;
      
      cancel.addActionListener(new ActionListener()
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent e)
         {
            dialog.setVisible(false);
            dialog.dispose();
         }
      });
      
      save.addActionListener(new ActionListener()
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent e)
         {
            AdminUtils.busy(client);
            
            MenuDef def = new MenuDef();
            
            if (menuItemsEdit.isEditing())
               menuItemsEdit.getCellEditor().stopCellEditing();
            
            menuFromWidgets(def, mName, mDesc, menuItemsTmEdit);

            int rc = 0;
            try
            {
               rc = majic.menuDbHelper(def, isEdit ? 
                                            MajicExtExports.MODE_EDIT : 
                                            MajicExtExports.MODE_ADD);
            }
            catch (Exception v)
            {
               client.beep("ERROR: " + v.getMessage());
               return;
            }
            finally
            {
               AdminUtils.idle(client);
            }
            
            if (rc == MajicExtExports.RC_OK)
               client.ok();
            else
            {
               String msg = "ERROR: Unable to update menu.";
               
               if (rc == MajicExtExports.RC_ALREADY_EXISTS)
               {
                  client.beep("ERROR: Menu already exists");
                  return;
               }
               else if (rc == MajicExtExports.RC_CANT_ADD)
                  msg = "ERROR: Unable to add menu";
               
               client.beep(msg.toString());
            }
            
            dialog.setVisible(false);
            dialog.dispose();
            populateMenus();
         }
      });
      
      dialog.setVisible(true);
   }

   /**
    * Prepare list of functions for menu editor.
    * 
    * @return   array of short function descriptors.
    */
   private ArrayList<String> extractFuncList()
   {
      if (funcCache != null && funcCache.size() > 0)
         return funcCache;
      
      if (functionTm.getRowCount() == 0)
      {
         populateFunctions();
      }

      if (funcCache == null)
         funcCache = new ArrayList<String>();
      
      for(int i = 0; i < functionTm.getRowCount(); i++)
      {
         String val = ((String) functionTm.getValueAt(i, 1)).trim();

         if (val.length() == 0)
            continue;

         funcCache.add(val);
      }

      Collections.sort(funcCache);
      
      return funcCache;
   }
   
   /**
    * Copy content of widgets into provided {@link MenuDef} instance.
    * 
    * @param    def
    * @param    desc
    * @param    name
    * @param    model
    */
   protected void menuFromWidgets(MenuDef def, 
                                  JTextField name,
                                  JTextField desc,
                                  TaggedDataTableModel model)
   {
      def.setName(name.getText());
      def.setDescription(desc.getText());

      for (int i = 0; i < model.getRowCount(); i++)
      {
         def.setHotKey(i, (String) model.getValueAt(i, 0));
         def.setHotKey(i + MenuDef.VECTOR_SIZE /2 , (String) model.getValueAt(i, 3));
         
         def.setEntryDesc(i, (String) model.getValueAt(i, 1));
         def.setEntryDesc(i + MenuDef.VECTOR_SIZE /2 , (String) model.getValueAt(i, 4));

         def.setFuncList(i, (String) model.getValueAt(i, 2));
         def.setFuncList(i + MenuDef.VECTOR_SIZE /2 , (String) model.getValueAt(i, 5));
      }
   }

   /**
    * Sum vector of integers.
    * 
    * @param    data
    * 
    * @return   sum of all elements of provided vector.
    */
   private static int sum(int... data)
   {
      int sum = 0;
      
      for (int i = 0; i < data.length; i++)
      {
         sum += data[i];
      }
      
      return sum;
   }

   /**
    * Add or update printer definition.
    * 
    * @param    values
    *           Values to update.
    */
   private void addOrUpdatePrinter(String ... values)
   {
      final boolean isEdit = values != null && values.length > 0;
      JPanel pnlInput  = new JPanel();
      
      GroupLayout layout = new GroupLayout(pnlInput);
      pnlInput.setLayout(layout);
      layout.setAutoCreateGaps(true);
      layout.setAutoCreateContainerGaps(true);
      
      JLabel lpName   = new JLabel("Printer Name");
      JLabel lpType   = new JLabel("Printer Type");
      JLabel lInFile  = new JLabel("Input File");
      JLabel lOutFile = new JLabel("Output File");
      JLabel lSString = new JLabel("Starting String");
      JLabel lEString = new JLabel("Ending String");
      JLabel lSProc   = new JLabel("Starting Procedure");
      JLabel lEProc   = new JLabel("Ending Procedure");

      final JTextField pName   = new JLimitedTextField(20);
      final JTextField pType   = new JLimitedTextField(12);
      final JTextField inFile  = new JLimitedTextField(1);
      final JTextField outFile = new JLimitedTextField(30);
      final JTextField sString = new JLimitedTextField(25);
      final JTextField eString = new JLimitedTextField(25);
      final JTextField sProc   = new JLimitedTextField(14);
      final JTextField eProc   = new JLimitedTextField(14);
      
      final JLabel[] lbl = new JLabel[] { lpName, lpType, lInFile, lOutFile, 
                                          lSString, lEString, lSProc, lEProc}; 

      final JTextField[] txt = new JTextField[] { pName, pType, inFile, outFile, 
                                                  sString, eString, sProc, eProc}; 
      
      if (isEdit)
         string2JComponent(values, txt);

      setMaximumSize(txt);
      
      pName.setEnabled(!isEdit);

      addInputs(pnlInput, txt);

      GroupLayout.SequentialGroup hGroup = layout.createSequentialGroup();
      
      // parallel group of labels
      addParallelGroup(hGroup, layout, lbl);
      // parallel group of input fields
      addParallelGroup(hGroup, layout, txt);
      
      layout.setHorizontalGroup(hGroup);
      layout.setVerticalGroup(newSequentialGroup(layout, lbl, txt));
      
      layout.linkSize(lbl);

      JPanel pnl2 = null;

      JButton cancel = new JButton("Cancel");
      JButton save   = new JButton("Save");
      
      pnl2 = AdminUtils.groupComponents(new Component[][] {{ cancel, save },});
      
      layout = (GroupLayout) pnl2.getLayout();
      layout.setAutoCreateContainerGaps(false);
      
      pnlInput.setBorder(new TitledBorder
                         (new EtchedBorder(EtchedBorder.RAISED), " Printer Definition "));
      
      JPanel groupComponents = AdminUtils.groupComponents(new Component[][] 
         { { pnlInput }, { pnl2 } } );
      
      final JDialog dialog = AdminUtils.modalDialog(client, 
                                                    "Printer Definition", 
                                                    groupComponents);

      cancel.addActionListener(new ActionListener()
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent e)
         {
            dialog.setVisible(false);
            dialog.dispose();
         }
      });
      
      save.addActionListener(new ActionListener()
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent e)
         {
            AdminUtils.busy(client);
            
            dialog.setVisible(false);
            
            TaggedName tn = new TaggedName(jComponent2string(txt));
            
            int rc = majic.printerDbHelper(tn, isEdit ? 
                                           MajicExtExports.MODE_EDIT : 
                                           MajicExtExports.MODE_ADD);

            AdminUtils.idle(client);
            
            if (rc == MajicExtExports.RC_OK)
               client.ok();
            else
            {
               String msg = "ERROR: Unable to update printer.";
               
               if (rc == MajicExtExports.RC_ALREADY_EXISTS)
               {
                  client.beep("ERROR: Printer already exists");
                  return;
               }
               else if (rc == MajicExtExports.RC_CANT_ADD)
                  msg = "ERROR: Unable to add printer";
               
               client.beep(msg);
            }
            
            dialog.dispose();
            populatePrinters();
         }
      });
      
      dialog.setVisible(true);
   }

   /**
    * Set preferred size for specified array of components.
    * 
    * @param    txt
    *           Array of components.
    */
   private static void setMaximumSize(final JComponent... txt)
   {
      for (JComponent field : txt)
      {
         field.setMaximumSize(field.getPreferredSize());
      }
   }

   /**
    * Add specified array of components to provided panel.
    * 
    * @param    pnlInput
    *           Panel to which components will be added.
    * @param    txt
    *           Array of components to add.
    */
   private static void addInputs(JPanel pnlInput, final JComponent... txt)
   {
      for (JComponent field : txt)
      {
         pnlInput.add(field);
      }
   }

   /**
    * Create sequential group from specified layout and components.
    * 
    * @param    layout
    *           Layout for which group is created.
    * @param    lbl
    *           Array of labels.
    * @param    txt
    *           Array of components.
    */
   private static GroupLayout.SequentialGroup newSequentialGroup(GroupLayout layout,
                                                                 JLabel[] lbl,
                                                                 JComponent[] txt)
   {
      GroupLayout.SequentialGroup vGroup = layout.createSequentialGroup();

      for (int i = 0; i < txt.length; i++)
      {
         vGroup.addGroup(layout.createParallelGroup(GroupLayout.Alignment.BASELINE)
                         .addComponent(lbl[i])
                         .addComponent(txt[i]));
      }

      return vGroup;
   }

   /**
    * Add components to specified group as a parallel group.
    * 
    * @param    hGroup
    *           Group to which components will be added.
    * @param    layout
    *           Layout for which parallel group is created.
    * @param    lbl
    *           Array of labels to add.
    */
   private static void addParallelGroup(SequentialGroup hGroup,
                                        GroupLayout layout, 
                                        JComponent[] lbl)
   {
      ParallelGroup group = layout.createParallelGroup();
      
      for (JComponent comp : lbl)
      {
         group.addComponent(comp);
      }
      
      hGroup.addGroup(group);
   }

   /**
    * Add or update function definition.
    * 
    * @param    values
    *           Values to update.
    */
   private void addOrUpdateFunction(final String ... values)
   {
      final boolean isEdit = values != null && values.length > 0;
      JPanel pnlInput  = new JPanel();
      
      GroupLayout layout = new GroupLayout(pnlInput);
      pnlInput.setLayout(layout);
      layout.setAutoCreateGaps(true);
      layout.setAutoCreateContainerGaps(true);
      
      final JLabel lfName   = new JLabel("Function Name");
      final JLabel lfDesc   = new JLabel("Function Description");
      final JLabel lfType   = new JLabel("Function Type");
      final JLabel lfID     = new JLabel("Function ID");
      final JLabel lfKeys   = new JLabel("Keystroke Access");
      final JLabel lfPrn    = new JLabel("Printer Name");
      final JLabel lfProg   = new JLabel("Program To Run");
      final JLabel lfDate   = new JLabel("Modifycation Date");
      final JLabel lfTrans  = new JLabel("Transaction allowed");
      final JLabel lfSymix  = new JLabel("Is Symix Function");
      final JLabel lfPrio   = new JLabel("Priority");
      final JLabel lfRun    = new JLabel("Is it invoked via RUN");
      final JLabel lfPermit = new JLabel("Only Administrator can change it");
      final JLabel lfTDesc  = new JLabel("Type Decription");
      final JLabel lfApp    = new JLabel("Approval Code");
      final JLabel lfDept   = new JLabel("Responsible Departments");
      final JLabel lfClass  = new JLabel("Function Class");

      final JTextField fName   = new JLimitedTextField(14, VALID_CHARS);
      final JTextField fDesc   = new JLimitedTextField(34);
      final JTextField fType   = new JLimitedTextField(10);
      final JTextField fID     = new JFormattedTextField(NumberFormat.getIntegerInstance());
      final JTextField fKeys   = new JLimitedTextField(10);
      final JTextField fPrn    = new JLimitedTextField(20);
      final JTextField fProg   = new JLimitedTextField(32);
      final JTextField fDate   = new JFormattedTextField(DateFormat.getDateInstance(DateFormat.SHORT, Locale.US));
      final JCheckBox  fTrans  = new JCheckBox();
      final JCheckBox  fSymix  = new JCheckBox();
      final JTextField fPrio   = new JFormattedTextField(NumberFormat.getIntegerInstance());
      final JCheckBox  fRun    = new JCheckBox();
      final JCheckBox  fPermit = new JCheckBox();
      final JTextField fTDesc  = new JLimitedTextField(12);
      final JTextField fApp    = new JLimitedTextField(4);
      final JTextField fDept    = new JLimitedTextField(16);
      final JRadioButton fClassRegular   = new JRadioButton("Regular (menu and permit)", true);
      final JRadioButton fClassPermit    = new JRadioButton("Permit only");

      final ButtonGroup fClassBtnGroup = new ButtonGroup();
      fClassBtnGroup.add(fClassRegular);
      fClassBtnGroup.add(fClassPermit);

      fPrio.setText("0");
      fID.setText("0");
      
      fID.setColumns(3);
      fDate.setColumns(10);
      fPrio.setColumns(3);

      final boolean[] hasAcl = new boolean[] {false};
      
      final JPanel pnlClassButtons = isEdit ?
          null :
          AdminUtils.groupComponents(new Component[][] {{fClassRegular, fClassPermit}});

      if (!isEdit)
      {
         GroupLayout pnlClassButtonsLayout = (GroupLayout) pnlClassButtons.getLayout();
         pnlClassButtonsLayout.setAutoCreateContainerGaps(false);
         pnlClassButtonsLayout.setAutoCreateGaps(false);
      }

      List<JComponent> txtList = new ArrayList<JComponent>();
      if (!isEdit)
      {
         txtList.add(pnlClassButtons);
      }

      txtList.addAll(Arrays.asList(
            fName, fDesc, fType, fID,
            fKeys, fPrn, fProg, fDate,
            fTrans, fSymix, fPrio, fRun,
            fPermit, fTDesc, fApp, fDept));

      final JComponent[] txt = txtList.toArray(new JComponent[0]);

      final String[] notesText     = new String[1];
      final String[] userSecurity  = new String[] {"*"};
      final String[] ownerSecurity = new String[] {"*"};
      final String[] userLevel     = new String[] {"0"};
      final String[] ownerLevel    = new String[] {"0"};
      final String functionOldDesc = isEdit ? values[1] : "";
      
      List<JComponent> lblList = new ArrayList<JComponent>();
      if (!isEdit)
      {
         lblList.add(lfClass);
      }

      lblList.addAll(Arrays.asList(
            lfName, lfDesc, lfType, lfID,
            lfKeys, lfPrn, lfProg, lfDate,
            lfTrans, lfSymix, lfPrio, lfRun,
            lfPermit, lfTDesc, lfApp, lfDept));

      final JLabel[] lbl = lblList.toArray(new JLabel[0]);

      JButton cancel = new JButton("Cancel");
      JButton save   = new JButton("Save");
      final JButton notes  = new JButton("Notes");
      final JButton access = new JButton("Security");

      final JComponent[] disabledForPermitFunc = new JComponent[]
            { lfName, lfType, lfID, lfKeys,
              lfPrn, lfProg, lfDate, lfTrans,
              lfSymix, lfPrio, lfRun, lfPermit,
              lfTDesc, lfApp, fName, fType, fID,
              fKeys, fPrn, fProg, fDate, fTrans,
              fSymix, fPrio, fRun, fPermit, fTDesc,
              fApp, notes, access
            };

      if (isEdit)
      {
         string2JComponent(values, txt);

         TaggedName fields = majic.retrieveFunctionComponents(fName.getText());
         
         if (fields.getName().length() > 0)
         {
            notesText[0]     = fields.get(5);
            userSecurity[0]  = fields.get(1) == null ? "*" : fields.get(1);
            ownerSecurity[0] = fields.get(2) == null ? "*" : fields.get(2);
            userLevel[0]     = fields.get(3) == null ? "0" : fields.get(3);
            ownerLevel[0]    = fields.get(4) == null ? "0" : fields.get(4);
         }
         
         Map<String, Integer> acls = majic.getACLs();
         
         if (acls != null && values != null && values.length > 2)
            hasAcl[0] = acls.containsKey(values[1]);

         // is permit function
         if (values[0].isEmpty())
         {
            setComponentsEnabled(false, disabledForPermitFunc);
         }
      }
      else
      {
         ((JFormattedTextField) fDate).setValue(new Date());
      }
      
      setMaximumSize(txt);

      fName.setEnabled(!isEdit);
      if (hasAcl[0])
         fDesc.setEnabled(!isEdit);

      addInputs(pnlInput, txt);

      GroupLayout.SequentialGroup hGroup = layout.createSequentialGroup();

      // parallel group of labels
      addParallelGroup(hGroup, layout, lbl);
      // parallel group of input fields
      addParallelGroup(hGroup, layout, txt);

      layout.setHorizontalGroup(hGroup);
      layout.setVerticalGroup(newSequentialGroup(layout, lbl, txt));

      layout.linkSize(lbl);

      JPanel pnl2 = null;

      pnl2 = AdminUtils.groupComponents(new Component[][] {
            { cancel, save, notes, new JLabel(), new JLabel(), access },
      });
      
      layout = (GroupLayout) pnl2.getLayout();
      layout.setAutoCreateContainerGaps(false);
      
      pnlInput.setBorder(new TitledBorder
                         (new EtchedBorder(EtchedBorder.RAISED), " Function Definition "));
      
      JPanel groupComponents = AdminUtils.groupComponents(
            new Component[][] { {pnlInput}, {pnl2} });

      final JDialog dialog = AdminUtils.modalDialog(client,
                                                    "Function Definition", 
                                                    groupComponents);
      
      cancel.addActionListener(new ActionListener()
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent e)
         {
            dialog.setVisible(false);
            dialog.dispose();
         }
      });
      
      save.addActionListener(new ActionListener()
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent e)
         {
            final boolean isPermitFunction = isEdit ? values[0].isEmpty() :
                                                fClassPermit.isSelected();

            if (!isPermitFunction)
            {
               // validation
               if (fType.getText().trim().toLowerCase().length() == 0)
               {
                  client.beep("ERROR: Function type must be defined");
                  return;
               }
            
               if (fName.getText().trim().toLowerCase().startsWith("text"))
               {
                  fProg.setText("");
               }
            
               if (!majic.validateKeyAccess(fName.getText(), fKeys.getText(), userLevel[0]))
               {
                  client.beep("WARNING: Function keys are already used");
               }
            
               String prog = fProg.getText().trim();
            
               prog = prog.replace('\\', '/');

               while (prog.endsWith("/"))
                  prog = prog.substring(0, prog.length() - 1);
            
               while (prog.startsWith("./"))
                  prog = prog.substring(2);
            
               if (prog.startsWith(".."))
               {
                  client.beep("ERROR: Paths pointing to parent directories are not allowed");
                  return;
               }
            
               fProg.setText(prog);
            }

            AdminUtils.busy(client);

            List<String> valuesList = new ArrayList<String>(Arrays.asList(jComponent2string(txt)));
            valuesList.add(notesText[0]);
            valuesList.add(userSecurity[0]);
            valuesList.add(ownerSecurity[0]);
            valuesList.add(userLevel[0]);
            valuesList.add(ownerLevel[0]);
            valuesList.add(isPermitFunction ? "permit" : "regular");
            valuesList.add(functionOldDesc);

            int rc = majic.functionDbHelper(new TaggedName(valuesList.toArray(new String[0])),
                                            isEdit ?
                                               MajicExtExports.MODE_EDIT :
                                               MajicExtExports.MODE_ADD);

            AdminUtils.idle(client);

            if (rc == MajicExtExports.RC_OK)
               client.ok();
            else
            {
               String msg = "ERROR: Unable to update function.";
               
               if (rc == MajicExtExports.RC_ALREADY_EXISTS)
               {
                  client.beep("ERROR: Function already exists");
                  return;
               }
               
               if (rc == MajicExtExports.RC_CANT_ADD)
                  msg = "ERROR: Unable to add function";
               
               client.beep(msg);
            }
            
            dialog.setVisible(false);
            dialog.dispose();
            populateFunctions();
         }
      });
      
      notes.addActionListener(new ActionListener()
      {
         @Override
         public void actionPerformed(ActionEvent e)
         {
            editNotes(notesText);
         }
      });

      access.addActionListener(new ActionListener()
      {
         @Override
         public void actionPerformed(ActionEvent e)
         {
            editSecurity(userSecurity, ownerSecurity, userLevel, ownerLevel);
         }
         
      });

      fClassPermit.addActionListener(new ActionListener()
      {
         @Override
         public void actionPerformed(ActionEvent e)
         {
            setComponentsEnabled(false, disabledForPermitFunc);
         }
      });

      fClassRegular.addActionListener(new ActionListener()
      {
         @Override
         public void actionPerformed(ActionEvent e)
         {
            setComponentsEnabled(true, disabledForPermitFunc);
         }
      });

      dialog.setVisible(true);
   }

   /**
    * Enables/disables the given set of components.
    *
    * @param enable
    *        <code>true</code> if the given components should be enabled,
    *        <code>false</code> if the given components should be disabled. 
    * @param components
    *        The set of components to enable/disable.   
    */
   private void setComponentsEnabled(boolean enable, JComponent[] components)
   {
      for (JComponent component : components)
      {
         component.setEnabled(enable);
      }
   }

   /**
    * Edit function notes.
    * 
    * @param    notesText
    */
   private void editNotes(final String[] notesText)
   {
      //edit notes
      JPanel pnlInput  = new JPanel();
      
      JPanel pnl2 = null;

      JButton cancel = new JButton("Cancel");
      JButton save   = new JButton("Save");
      
      final JTextArea text = new JTextArea(15, 40);

      text.setMaximumSize(text.getPreferredSize());
      text.setLineWrap(true);
      text.setDocument(new LimitedDocument(15 * 40, null));
      text.setText(notesText[0]);
      
      pnlInput.setLayout(new BorderLayout());
      
      pnlInput.add(text, BorderLayout.CENTER);
      
      pnl2 = AdminUtils.groupComponents(new Component[][] {
            { cancel, save},
      });
      GroupLayout layout = (GroupLayout) pnl2.getLayout();
      layout.setAutoCreateContainerGaps(false);
      
      pnlInput.setBorder(new TitledBorder
                         (new EtchedBorder(EtchedBorder.RAISED), " Notes "));
      
      JPanel groupComponents = AdminUtils.groupComponents(new Component[][] 
         { { pnlInput }, { pnl2 } } );
      
      final JDialog dialog = AdminUtils.modalDialog(client, 
                                                    "Function Notes", 
                                                    groupComponents);
      
      cancel.addActionListener(new ActionListener()
      {
         @Override
         public void actionPerformed(ActionEvent e)
         {
            dialog.setVisible(false);
            dialog.dispose();
         }
      });
      
      save.addActionListener(new ActionListener()
      {
         @Override
         public void actionPerformed(ActionEvent e)
         {
            notesText[0] = text.getText();
            
            dialog.setVisible(false);
            dialog.dispose();
         }
      });
      
      dialog.setVisible(true);
   }

   /**
    * Edit function security.
    */
   private void editSecurity(final String[] userSecurity, 
                             final String[] ownerSecurity,
                             final String[] userLevel,
                             final String[] ownerLevel)
   {
      //edit notes
      JPanel pnlInput  = new JPanel();
      
      JPanel pnl2 = null;

      JButton cancel = new JButton("Cancel");
      JButton save   = new JButton("Save");
  
      GroupLayout layout = new GroupLayout(pnlInput);
      pnlInput.setLayout(layout);
      layout.setAutoCreateGaps(true);
      layout.setAutoCreateContainerGaps(true);

      JLabel lfUserSec   = new JLabel("User Srcurity");
      JLabel lfUserLevel = new JLabel("User Level");
      JLabel lfOwnSec    = new JLabel("Owner Security");
      JLabel lfOwnLevel  = new JLabel("Owner Level");

      final JTextField fUserSec   = new JLimitedTextField(40, 255);
      final JTextField fUserLevel = new JFormattedTextField(NumberFormat.getIntegerInstance());
      final JTextField fOwnSec    = new JLimitedTextField(40, 255);
      final JTextField fOwnLevel  = new JFormattedTextField(NumberFormat.getIntegerInstance());
      
      fUserLevel.setColumns(2);
      //fUserLevel.setDocument(new LimitedDocument(2));
      fOwnLevel.setColumns(2);
      //fOwnLevel.setDocument(new LimitedDocument(2));
      fUserSec.getHorizontalVisibility().setRangeProperties(0, 255, 0, 255, false);
      fOwnSec.getHorizontalVisibility().setRangeProperties(0, 255, 0, 255, false);
      
      fUserLevel.setText(userLevel[0]);
      fOwnLevel.setText(ownerLevel[0]);      
      fUserSec.setText(userSecurity[0]);
      fOwnSec.setText(ownerSecurity[0]);
      
      final JComponent[] txt = new JComponent[] {fUserSec, fUserLevel, fOwnSec,fOwnLevel}; 
      final JLabel[] lbl = new JLabel[] {lfUserSec, lfUserLevel, lfOwnSec, lfOwnLevel}; 

      setMaximumSize(txt);

      addInputs(pnlInput, txt);

      GroupLayout.SequentialGroup hGroup = layout.createSequentialGroup();

      // parallel group of labels
      addParallelGroup(hGroup, layout, lbl);
      // parallel group of input fields
      addParallelGroup(hGroup, layout, txt);

      layout.setHorizontalGroup(hGroup);
      layout.setVerticalGroup(newSequentialGroup(layout, lbl, txt));

      layout.linkSize(lbl);
      
      pnl2 = AdminUtils.groupComponents(new Component[][] { { cancel, save}, });
      
      pnlInput.setBorder(new TitledBorder
                         (new EtchedBorder(EtchedBorder.RAISED), " Notes "));

      layout = (GroupLayout) pnl2.getLayout();
      layout.setAutoCreateContainerGaps(false);
      
      pnlInput.setBorder(new TitledBorder
                         (new EtchedBorder(EtchedBorder.RAISED), " Function Security "));
      
      JPanel groupComponents = AdminUtils.groupComponents(new Component[][] 
         { { pnlInput }, { pnl2 } } );
      
      final JDialog dialog = AdminUtils.modalDialog(client, 
                                                    "Function Security", 
                                                    groupComponents);
      
      cancel.addActionListener(new ActionListener()
      {
         @Override
         public void actionPerformed(ActionEvent e)
         {
            dialog.setVisible(false);
            dialog.dispose();
         }
      });
      
      save.addActionListener(new ActionListener()
      {
         @Override
         public void actionPerformed(ActionEvent e)
         {
            userLevel[0]     = fUserLevel.getText();
            ownerLevel[0]    = fOwnLevel.getText();
            userSecurity[0]  = fUserSec.getText();
            ownerSecurity[0] = fOwnSec.getText();
            
            dialog.setVisible(false);
            dialog.dispose();
         }
      });
      
      dialog.setVisible(true);
   }

   /**
    * Scattering text from string array to input components fields.
    * 
    * @param    values
    *           Array of values to scatter.
    * @param    txt
    *           Array of text fields.
    */
   private static void string2JComponent(String[] values, JComponent[] txt)
   {
      for (int i = 0; i < values.length && i < txt.length; i++)
      {
         if (txt[i] instanceof JTextField)
            ((JTextField) txt[i]).setText(values[i]);
         if (txt[i] instanceof JCheckBox)
            ((JCheckBox) txt[i]).getModel().setSelected(values[i].equalsIgnoreCase("yes"));
      }
   }

   /**
    * Extract table row values.
    * 
    * @param    model
    *           Table model to extract values from.
    * @param    ndx
    *           Index of row to extract.
    * @param    startCol
    *           Column index in row to start extracting data.
    * 
    * @return   array of extracted values for first selected row or
    *           <code>null</code> if no row is selected.
    */
   private String[] getRowValues(TableModel model, int ndx, int startCol)
   {
      if (ndx < 0 || startCol < 0)
         return null;
      
      String[] values = new String[model.getColumnCount() - startCol];
      
      for (int i = startCol; i < (values.length + startCol); i++)
      {
         values[i - startCol] = (String) model.getValueAt(ndx, i);
      }
      return values;
   }

   /**
    * Gather strings from input components.
    * 
    * @param    txt
    *           Array of text fields.
    * 
    * @return   Array of collected strings.
    */
   private static String[] jComponent2string(JComponent[] txt)
   {
      ArrayList<String> list = new ArrayList<String>();
      
      for (JComponent field : txt)
      {
         if (field instanceof JTextField)
            list.add(((JTextField) field).getText());
         if (field instanceof JCheckBox)
            list.add(((JCheckBox) field).getModel().isSelected() ? "yes" : "no");
      }
      
      return list.toArray(new String[list.size()]);
   }
   
   /**
    * Inserts custom major and minor menu items.
    * The inserted menu items have to have the event listeners attached
    * for the menu items to have effect.
    * <p>
    * This implementation inserts: 
    * <ul>
    *   <li>"MAJIC" major menu after "Access Control" menu with the "Functions"
    *        menu item
    *   <li>"MAJIC" sub-menu into "Runtime Setting" major menu, Per Server
    * </ul>
    *  
    * @param    menu
    *           the main menu list
    */
   public void insertCustomMenuItems(List<JMenuItem> menu)
   {
      //menu.add(new JMenuItem("hello!!!"));
      // insert the minor menu items
      // get the third menu - Runtime Settings
      JMenu rts = (JMenu)menu.get(3);
 
      // get the first sub-menu - Per Server
      psr = (JMenu)rts.getItem(0);
      
      // append to the end of sub-menu
      psrCount = psr.getItemCount();
      comp = new JMenuItem("Company Name");
      
      if (psrCount > 0)
      {
         psr.addSeparator();
      }
      
      psr.add(comp);
      
      // event handler for the Company Name item
      comp.addActionListener(new ActionListener()
      {
         @Override
         @SuppressWarnings("unused")
         public void actionPerformed(ActionEvent evt) 
         {
            companyName();
         }
      });
      
      // insert the major MAJIC menu 
      majicMenu = new JMenu("MAJIC");
      menu.add(MAJIC_MENU_POS, majicMenu);
      
      majicMenu.add(xprint  = new JMenuItem("Printer Maintenance"));
      majicMenu.add(fun     = new JMenuItem("System Function Maintenance"));
      
      majicMenu.add(xmbuild = new JMenuItem("Panel Menu Builder"));
      
      // event handler for the Printer Maintenance item
      xprint.addActionListener(new ActionListener()
      {
         @Override
         @SuppressWarnings("unused")
         public void actionPerformed(ActionEvent evt) 
         {
            majicPrinters();
         }
      });

      // event handler for the Functions item
      fun.addActionListener(new ActionListener()
      {
         @Override
         @SuppressWarnings("unused")
         public void actionPerformed(ActionEvent evt) 
         {
            majicFunctions();
         }
      });

      if (xmbuild != null)
      {
         // event handler for the Menu Builder item
         xmbuild.addActionListener(new ActionListener()
         {
            @Override
            @SuppressWarnings("unused")
            public void actionPerformed(ActionEvent evt) 
            {
               majicMenuBuilder();
            }
         });
      }

      // insert the MAJIC Reports menu 
      majicReportsMenu = new JMenu("MAJIC Reports");
      menu.add(MAJIC_MENU_POS, majicReportsMenu);
      
      // User Auth Form report
      repUAFMenuItem =
         new JMenuItem(UserAuthFormReport.REPORT_TITLE);
      majicReportsMenu.add(repUAFMenuItem);
      repUserAuthForm = new UserAuthFormReport(client, majic);
      
      // event handler for the User Auth Form report
      repUAFMenuItem.addActionListener(new ActionListener()
      {
         @Override
         @SuppressWarnings("unused")
         public void actionPerformed(ActionEvent evt) 
         {
            repUserAuthForm.showDialog();
         }
      });

      // User Auth By Function report
      JMenuItem repUABFMenuItem =
         new JMenuItem(UserAuthByFuncReport.REPORT_TITLE);
      majicReportsMenu.add(repUABFMenuItem);
      repUserAuthByFunc = new UserAuthByFuncReport(client, majic);

      // event handler for the User Auth By Function report
      repUABFMenuItem.addActionListener(new ActionListener()
      {
         @Override
         @SuppressWarnings("unused")
         public void actionPerformed(ActionEvent evt)
         {
            repUserAuthByFunc.showDialog();
         }
      });

      // User List report
      JMenuItem repULMenuItem =
         new JMenuItem(UserListReport.REPORT_TITLE);
      majicReportsMenu.add(repULMenuItem);
      repUserList = new UserListReport(client, majic);

      // event handler for the User Auth Form report
      repULMenuItem.addActionListener(new ActionListener()
      {
         @Override
         @SuppressWarnings("unused")
         public void actionPerformed(ActionEvent evt)
         {
            repUserList.showDialog();
         }
      });
   }

   /**
    * Conditionally removes custom major and minor menu items. This method
    * should remove all menu items that were inserted by the last call to
    * {@link #insertCustomMenuItems}
    * 
    * @param    menu
    *           the main menu list
    */
   public void removeCustomMenuItems(List<JMenuItem> menu)
   {
      // check for previous insertions 
      JMenuItem mi = menu.get(MAJIC_MENU_POS);

      if (!mi.equals(majicReportsMenu))
      {
         // no custom menus
         return;
      }
      
      // remove the report menu
      menu.remove(MAJIC_MENU_POS);
      
      // check for previous insertions 
      mi = menu.get(MAJIC_MENU_POS);
      if (mi.equals(majicMenu))
      {
         // remove the major menu 
         menu.remove(MAJIC_MENU_POS);
      }
      
      // remove the company name
      if (psrCount > 0)
      {
         // remove separator
         psr.remove(psrCount);
      }

      // remove Company Name menu item
      psr.remove(psrCount);
   }
   
   /**
    * MAJIC Panel Menu Builder screen.
    */
   private void majicMenuBuilder()
   {
      populateMenus();
      client.show(SCREEN_MENU_BUILDER, TITLE_MENU_BUILDER);
   }
   
   /**
    * MAJIC Functions management screen.
    */
   private void majicFunctions()
   {
      populateFunctions();
      client.show(SCREEN_FUNCTION_MAINTENANCE, TITLE_FUNCTION_MAINTENANCE);
   }

   /**
    * MAJIC Printer Maintenance management screen.
    */
   private void majicPrinters()
   {
      populatePrinters();
      client.show(SCREEN_PRINTER_MAINTENANCE, TITLE_PRINTER_MAINTENANCE);
   }
   
   /**
    * Company Name editor.
    */
   private void companyName()
   {
      AdminExports adm = client.getExports();
      
      // get servers where majic/company is defined
      String[] servers = adm.listServers("majic/company");
      if (servers == null || servers.length == 0)
      {
         client.beep("found no servers with MAJIC Company Name");
         return;
      }
      
      // identify instance to edit: 1st specific, 2nd default
      String ownName = adm.getServerName();
      int idx = AdminUtils.isListed(ownName, servers);
      
      if (idx == -1)
      {
         idx = AdminUtils.isListed("", servers);
      }
      
      if (idx == -1)
      {
         client.beep("target server has no MAJIC Company Name");
         return;
      }
      
      // get the current value
      String company = majic.getCompanyName(servers[idx]);
      
      // create an editing panel
      JLabel     cnl = new JLabel("Company Name");
      JTextField cnf = new JTextField(40);
      JButton    cnc = new JButton("Cancel");      
      JButton    cns = new JButton("Save");      
      
      cnf.setText(company);

      Component[][] comp = new Component[2][2];
      comp[0][0] = cnl;
      comp[0][1] = cnf;
      comp[1][0] = cnc;
      comp[1][1] = cns;
      
      JPanel panel = AdminUtils.groupComponents(comp, false);
      
      // create a modal editor dialog
      dlg = AdminUtils.modalDialog(client, "Company Name", panel);
      
      // attach event handler for the Cancel button
      cnc.addActionListener(new ActionListener()
      {
         @Override
         @SuppressWarnings("unused")
         public void actionPerformed(ActionEvent evt) 
         {
            dlg.setVisible(false);
            dlg = null;
         }
      });
      
      // attach event handler for the Save button
      cns.addActionListener(new ActionListener()
      {
         @Override
         @SuppressWarnings("unused")
         public void actionPerformed(ActionEvent evt) 
         {
            dlg.setVisible(false);
         }
      });
      
      // display the dialog
      dlg.setVisible(true);
      
      // check the results
      if (dlg == null)
      {
         // cancelled
         return;
      }
      
      company = cnf.getText();
      
      // save the new value
      if (!majic.setCompanyName(servers[idx], company))
      {
         client.beep("saving company name failed");
      }
   }

   /**
    * Handle press of the "Find" button on the user search screen. 
    *
    * @param findScreen
    *        Backing find screen object.
    */
   private void handlePerformUserFind(IFindAccountScreen findScreen)
   {
      String text = findUserInput.getText().trim();

      Integer empNum = null;
      try
      {
         empNum = Integer.parseInt(text);
      }
      catch (NumberFormatException e)
      {
         // suppress
      }

      String user = null;
      if (empNum != null)
      {
         user = majic.getUserNameByEmployeeNumber(empNum);
      }

      findScreen.findAccountByName(user != null ? user : text);
   }

   /**
    * Defines a helper interface, used in the browser dialog.
    */
   private interface IBrowser
   {
      /**
       * Using the specified id and name filters, retrieve the records in
       * the specified page.
       * 
       * @param   idFilter
       *          current ID filter
       * @param   nameFilter
       *          current NAME filter
       * @param   page
       *          next page
       * @param   pageSize
       *          current page size 
       * 
       * @return  an array with the current results.
       */
      public TaggedName[] findRecords(String idFilter, 
                                      String nameFilter, 
                                      int    page, 
                                      int    pageSize);

      /**
       * Return the id widget associated with the browse dialog.
       * 
       * @return   the id widget
       */
      public JTextField getIdField();

      /**
       * Return the description widget associated with the browse dialog.
       * 
       * @return   the description widget
       */
      public JLabel getDescField();
      
      /**
       * Get the title of the browse dialog.
       * 
       * @return   the browser title
       */
      public String getTitle();
      
      /**
       * Get the text for the first column
       * 
       * @return   text for the first column
       */
      public String getFirstColLabel();
      
      /**
       * Get the text for the second column
       * 
       * @return   text for the second column
       */
      public String getSecondColLabel();
      
      /**
       * Handle any custom actions once a row is chosen in the browse.
       * 
       * @param    id
       *           the id of the chosen row
       * @param    desc
       *           the description for the chosen row
       */
      public void handleChoose(String id, String desc);
   }

   /**
    * Defines employee browser.
    */
   private class EmployeeBrowser
   implements IBrowser
   {
      /**
       * Using the specified id and name filters, retrieve the employees in
       * the specified page.
       * 
       * @param    idFilter
       *           current ID filter
       * @param    nameFilter
       *           current NAME filter
       * @param    page
       *           next page
       * @param    pageSize
       *           current page size
       * 
       * @return   an array with the current results.
       */
      public TaggedName[] findRecords(String idFilter, 
                                      String nameFilter, 
                                      int page, 
                                      int pageSize)
      {
         return majic.listEmployees(idFilter, nameFilter, page, pageSize);
      }
      
      /**
       * Return the id widget associated with the browse dialog.
       * 
       * @return   the id widget
       */
      public JTextField getIdField()
      {
         return t1fEId;
      }
      
      /**
       * Return the description widget associated with the browse dialog.
       * 
       * @return   the description widget
       */
      public JLabel getDescField()
      {
         return t1lEName;
      }
      
      /**
       * Get the title of the browse dialog.
       * 
       * @return   the browser title
       */
      public String getTitle()
      {
         return "Select Employee";
      }
      
      /**
       * Get the text for the first column
       * 
       * @return   text for the first column: Employee #
       */
      public String getFirstColLabel()
      {
         return "Employee #";
      }
      
      /**
       * Get the text for the second column
       * 
       * @return   text for the second column: Person
       */
      public String getSecondColLabel()
      {
         return "Person";
      }

      /**
       * Handle any custom actions once a row is chosen in the browse.
       */
      public void handleChoose(String id, String desc)
      {
         /*
         // get PERSON field, if this is the user-account dialog
         Component c = ias.getWidgetByName(AdminUserAccounts.WIDGET_PERSON);
         if (c != null)
         {
            desc = desc.trim();

            // found the person widget, copy the person's name
            if (c instanceof JTextComponent)
            {
               if (desc.length() > 0)
               {
                  ((JTextComponent) c).setText(desc);
               }
               
               c.setEnabled(desc.length() == 0 || "0".equals(id));
            }
         }
         */
      }
   }
   
   /**
    * Defines customer browser.
    */
   private class CustomerBrowser
   implements IBrowser
   {
      /**
       * Using the specified id and name filters, retrieve the customers in
       * the specified page.
       * 
       * @param    idFilter
       *           current ID filter
       * @param    nameFilter
       *           current NAME filter
       * @param    page
       *           next page
       * @param    pageSize
       *           current page size
       * 
       * @return   an array with the current results.
       */
      public TaggedName[] findRecords(String idFilter, 
            String nameFilter, 
            int page, 
            int pageSize)
      {
         return majic.listCustomers(idFilter, nameFilter, page, pageSize);
      }
      
      /**
       * Return the id widget associated with the browse dialog.
       * 
       * @return   the id widget
       */
      public JTextField getIdField()
      {
         return t1fCId;
      }
      
      /**
       * Return the description widget associated with the browse dialog.
       * 
       * @return   the description widget
       */
      public JLabel getDescField()
      {
         return t1lCName;
      }
      
      /**
       * Get the title of the browse dialog.
       * 
       * @return   the browser title
       */
      public String getTitle()
      {
         return "Select Customer";
      }
      
      /**
       * Get the text for the first column
       * 
       * @return   text for the first column: Customer #
       */
      public String getFirstColLabel()
      {
         return "Customer #";
      }
      
      /**
       * Get the text for the second column
       * 
       * @return   text for the second column: Name
       */
      public String getSecondColLabel()
      {
         return "Name";
      }
      
      /**
       * Handle any custom actions once a row is chosen in the browse.
       * <p>
       * This is a no-op.
       * 
       * @param    id
       * @param    desc
       */
      public void handleChoose(String id, String desc)
      {
         // no-op
      }
   }
   
   /**
    * Defines Printer browser.
    */
   private class PrinterBrowser
   implements IBrowser
   {
      /**
       * Using the specified id and name filters, retrieve the printers in the
       * specified page.
       * 
       * @param    idFilter
       *           current ID filter
       * @param    nameFilter
       *           current NAME filter
       * @param    page
       *           next page
       * @param    pageSize
       *           current page size
       * 
       * @return   an array with the current results.
       */
      public TaggedName[] findRecords(String idFilter, 
            String nameFilter, 
            int page, 
            int pageSize)
      {
         return majic.listPrinters(idFilter, nameFilter, page, pageSize);
      }
      
      /**
       * Return the id widget associated with the browse dialog.
       * 
       * @return   the id widget
       */
      public JTextField getIdField()
      {
         return t3fPrId;
      }
      
      /**
       * Return the description widget associated with the browse dialog.
       * 
       * @return   the description widget
       */
      public JLabel getDescField()
      {
         return t3lPrName;
      }
      
      /**
       * Get the title of the browse dialog.
       * 
       * @return   the browser title
       */
      public String getTitle()
      {
         return "Select Printer";
      }
      
      /**
       * Get the text for the first column
       * 
       * @return   text for the first column: Printer
       */
      public String getFirstColLabel()
      {
         return "Printer";
      }
      
      /**
       * Get the text for the second column
       * 
       * @return   text for the second column: Description
       */
      public String getSecondColLabel()
      {
         return "Description";
      }

      /**
       * Handle any custom actions once a row is chosen in the browse.
       * <p>
       * This is a no-op.
       * 
       * @param    id
       * @param    desc
       */
      public void handleChoose(String id, String desc)
      {
         // no-op
      }
   }

   /**
    * Make plural from noun.
    * 
    * @param    count
    *           The number of things.
    * @param    string
    *           Source string.
    *           
    * @return   Input string if count is 1 or plural form of noun if count is
    *           greater than 1.
    */
   private static String makePlural(int count, String string)
   {
      return string + (count > 1 ? "s":"");
   }
   
   /**
    * Defines Entry point browser.
    */
   private class EntryPointBrowser
   implements IBrowser
   {
      /**
       * Using the specified id and name filters, retrieve the entry-points in
       * the specified page.
       * 
       * @param   idFilter
       *          current ID filter
       * @param   nameFilter
       *          current NAME filter
       * @param   page
       *          next page
       * @param   pageSize
       *          current page size 
       * 
       * @return  an array with the current results.
       */
      public TaggedName[] findRecords(String idFilter, 
            String nameFilter, 
            int page, 
            int pageSize)
      {
         return majic.listEntryPoints(idFilter, 
                                          nameFilter, 
                                          page, 
                                          pageSize);
      }
      
      /**
       * Return the id widget associated with the browse dialog.
       * 
       * @return   the id widget
       */
      public JTextField getIdField()
      {
         return t2fEpId;
      }
      
      /**
       * Return the description widget associated with the browse dialog.
       * 
       * @return   the description widget
       */
      public JLabel getDescField()
      {
         return t2lEpName;
      }
      
      /**
       * Get the title of the browse dialog.
       * 
       * @return   the browser title
       */
      public String getTitle()
      {
         return "Select Entry Point";
      }
      
      /**
       * Get the text for the first column
       * 
       * @return   text for the first column: Function
       */
      public String getFirstColLabel()
      {
         return "Function";
      }
      
      /**
       * Get the text for the second column
       * 
       * @return   text for the second column: Description
       */
      public String getSecondColLabel()
      {
         return "Description";
      }

      /**
       * Handle any custom actions once a row is chosen in the browse.
       * <p>
       * This is a no-op.
       * 
       * @param    id
       * @param    desc
       */
      public void handleChoose(String id, String desc)
      {
         // no-op
      }
   }
   
   /**
    * Simple extension of {@link PlainDocument} which limits user input to
    * specified length.
    */
   static class LimitedDocument
   extends PlainDocument
   {
      /** Text length limit. */
      private final int limit;

      /** Valid chars */
      private final String validChars;
      
      /**
       * Constructor.
       * 
       * @param limit
       */
      LimitedDocument(int limit, String validChars)
      {
         super();
         this.validChars = validChars;
         this.limit = limit;
      }

      /**
       * @see javax.swing.text.PlainDocument#insertString(int, java.lang.String, javax.swing.text.AttributeSet)
       */
      @Override
      public void insertString(int offset, String str, AttributeSet attr)
      throws BadLocationException
      {
         if (str == null)
            return;

         int length  = getLength();
         int maxSize = limit - length;

         if (maxSize > 0 && str.length() > maxSize)
            str = str.substring(0, maxSize);

         if ((length + str.length()) <= limit)
         {
            super.insertString(offset, str, attr);
         }
      }
   }
   
   /**
    * Extended {@link JTextField} which limits user input length.
    */
   static class JLimitedTextField
   extends JTextField
   {
      /**
       * Construct field which has input limit and preferred visible size
       * identical.
       * 
       * @param    limit
       */
      JLimitedTextField(int limit)
      {
         this(limit, limit);
      }

      /**
       * Construct field which has input limit and preferred visible size
       * identical. Also this field limits accepted characters to specified
       * ones.
       * 
       * @param    limit
       */
      JLimitedTextField(int limit, String validChars)
      {
         this(limit, limit, validChars);
      }
      
      /**
       * Construct field which has input limit and preferred visible size
       * set to specified values.
       * 
       * @param visible
       * @param textLimit
       */
      JLimitedTextField(int visible, int textLimit)
      {
         this(visible, textLimit, null);
      }
      
      /**
       * Construct field which has input limit and preferred visible size set
       * to specified values. Also this field limits accepted characters to
       * specified ones.
       * 
       * @param    visible
       * @param    textLimit
       * @param    validChars
       */
      JLimitedTextField(int visible, int textLimit, String validChars)
      {
         super(visible);
         setDocument(new LimitedDocument(textLimit, validChars));
      }
   }

   abstract static class AbstractPrintRequest
   implements PrintRequest
   {
      /**
       * Queries the extended printing capability. If the extended capability
       * is not provided, the printExtended() method should not be called.
       *
       * @return <code>true</code> if the extended printing is implemented
       */
      @Override
      public boolean canPrintExtended()
      {
         return false;
      }

      /**
       * Performs printing of the current screen data.
       */
      @Override
      abstract public void print();

      /**
       * Performs extended printing of the current screen data.
       * This method will only be used if {@link #canPrintExtended} returns
       * <code>true</code>.
       */
      @Override
      public void printExtended()
      {
      }
   }
   
   /**
    * Print request handler for printer maintenance panel.
    */
   class PrinterPrinter
   extends AbstractPrintRequest
   {
      /**
       * Performs printing of the current screen data.
       */
      @Override
      public void print()
      {
         try
         {
            printers.print(PrintMode.FIT_WIDTH, 
                           new MessageFormat("Defined Printers"), 
                           new MessageFormat("page {0, number}"));
         }
         catch (PrinterException pe)
         {
            client.beep("printer exception" + pe);
         }
         catch (SecurityException se)
         {
            // nothing to do
         }
      }
   }
   
   /**
    * Print request handler for function maintenance panel.
    */
   class FunctionPrinter
   extends AbstractPrintRequest
   {
      /**
       * Performs printing of the current screen data.
       */
      @Override
      public void print()
      {
         try
         {
            functions.print(PrintMode.FIT_WIDTH, 
                            new MessageFormat("Defined Functions"), 
                            new MessageFormat("page {0, number}"));
         }
         catch (PrinterException pe)
         {
            client.beep("printer exception" + pe);
         }
         catch (SecurityException se)
         {
            // nothing to do
         }
      }
   }

   /**
    * Print request handler for menu builder panel.
    */
   public class MenuPrinter
   extends AbstractPrintRequest
   {
      /**
       * Performs printing of the current screen data.
       */
      @Override
      public void print()
      {
         boolean printItems = menus.getSelectedRowCount() > 0;
         String description = "";
         
         if (printItems)
         {
            // offering a choice 
            String quest = "Please refine your request";
            String[] choices = new String[] {"Print menu list", 
                                             "Print menu items for selected menu"};
            
            int n = AdminUtils.choiceDialog(client, "Printing Menus", quest,
                                            choices, "Print");
            if (n < 0)
            {
               // declined
               return;
            }
            
            if (n == 0)
               printItems = false;
            else
            {
               description = (String) menuListTm.getValueAt(menus.getSelectedRow(), 1); 
            }
         }
         
         try
         {
            if (printItems)
            {
               menuItems.print(PrintMode.FIT_WIDTH, 
                               new MessageFormat("Defined Menu Items for '" +
                                                 description +
                                                 "'"), 
                               new MessageFormat("page {0, number}"));
            }
            else
            {
               menus.print(PrintMode.FIT_WIDTH, 
                           new MessageFormat("Defined Menus"), 
                           new MessageFormat("page {0, number}"));
            }
         }
         catch (PrinterException pe)
         {
            client.beep("printer exception" + pe);
         }
         catch (SecurityException se)
         {
            // nothing to do
         }
      }
   }

   /**
    * ComboBox cell editor.
    */
   static class ComboBoxEditor 
   extends DefaultCellEditor 
   {
      /**
       * Constructor.
       * 
       * @param    items
       */
      public ComboBoxEditor(List<String> items) 
      {
         super(new JComboBox(items.toArray()));
         ((JComboBox)getComponent()).setEditable(true);
      }
   }

	@Override
	/**
    * Call to show client specific ACL
    * @param adminACL the admin ACL from where this is called
    */
   public void showStandaloneAcls(AdminACL adminACL)
   {
      adminACL.showByResourceType("majic-function");  
   }
}
